# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Networks for the PPO algorithm defined as recurrent cells."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tf.compat.v1 as tf

_MEAN_WEIGHTS_INITIALIZER = tf.contrib.layers.variance_scaling_initializer(factor=0.1)
_LOGSTD_INITIALIZER = tf.random_normal_initializer(-1, 1e-10)


class LinearGaussianPolicy(tf.contrib.rnn.RNNCell):
  """Indepent linear network with a tanh at the end for policy and feedforward network for the value.

  The policy network outputs the mean action and the log standard deviation
  is learned as indepent parameter vector.
  """

  def __init__(self,
               policy_layers,
               value_layers,
               action_size,
               mean_weights_initializer=_MEAN_WEIGHTS_INITIALIZER,
               logstd_initializer=_LOGSTD_INITIALIZER):
    self._policy_layers = policy_layers
    self._value_layers = value_layers
    self._action_size = action_size
    self._mean_weights_initializer = mean_weights_initializer
    self._logstd_initializer = logstd_initializer

  @property
  def state_size(self):
    unused_state_size = 1
    return unused_state_size

  @property
  def output_size(self):
    return (self._action_size, self._action_size, tf.TensorShape([]))

  def __call__(self, observation, state):
    with tf.variable_scope('policy'):
      x = tf.contrib.layers.flatten(observation)
      mean = tf.contrib.layers.fully_connected(x,
                                               self._action_size,
                                               tf.tanh,
                                               weights_initializer=self._mean_weights_initializer)
      logstd = tf.get_variable('logstd', mean.shape[1:], tf.float32, self._logstd_initializer)
      logstd = tf.tile(logstd[None, ...], [tf.shape(mean)[0]] + [1] * logstd.shape.ndims)
    with tf.variable_scope('value'):
      x = tf.contrib.layers.flatten(observation)
      for size in self._value_layers:
        x = tf.contrib.layers.fully_connected(x, size, tf.nn.relu)
      value = tf.contrib.layers.fully_connected(x, 1, None)[:, 0]
    return (mean, logstd, value), state


class ForwardGaussianPolicy(tf.contrib.rnn.RNNCell):
  """Independent feed forward networks for policy and value.

  The policy network outputs the mean action and the log standard deviation
  is learned as independent parameter vector.
  """

  def __init__(self,
               policy_layers,
               value_layers,
               action_size,
               mean_weights_initializer=_MEAN_WEIGHTS_INITIALIZER,
               logstd_initializer=_LOGSTD_INITIALIZER):
    self._policy_layers = policy_layers
    self._value_layers = value_layers
    self._action_size = action_size
    self._mean_weights_initializer = mean_weights_initializer
    self._logstd_initializer = logstd_initializer

  @property
  def state_size(self):
    unused_state_size = 1
    return unused_state_size

  @property
  def output_size(self):
    return (self._action_size, self._action_size, tf.TensorShape([]))

  def __call__(self, observation, state):
    with tf.variable_scope('policy'):
      x = tf.contrib.layers.flatten(observation)
      for size in self._policy_layers:
        x = tf.contrib.layers.fully_connected(x, size, tf.nn.relu)
      mean = tf.contrib.layers.fully_connected(x,
                                               self._action_size,
                                               tf.tanh,
                                               weights_initializer=self._mean_weights_initializer)
      logstd = tf.get_variable('logstd', mean.shape[1:], tf.float32, self._logstd_initializer)
      logstd = tf.tile(logstd[None, ...], [tf.shape(mean)[0]] + [1] * logstd.shape.ndims)
    with tf.variable_scope('value'):
      x = tf.contrib.layers.flatten(observation)
      for size in self._value_layers:
        x = tf.contrib.layers.fully_connected(x, size, tf.nn.relu)
      value = tf.contrib.layers.fully_connected(x, 1, None)[:, 0]
    return (mean, logstd, value), state


class RecurrentGaussianPolicy(tf.contrib.rnn.RNNCell):
  """Independent recurrent policy and feed forward value networks.

  The policy network outputs the mean action and the log standard deviation
  is learned as independent parameter vector. The last policy layer is recurrent
  and uses a GRU cell.
  """

  def __init__(self,
               policy_layers,
               value_layers,
               action_size,
               mean_weights_initializer=_MEAN_WEIGHTS_INITIALIZER,
               logstd_initializer=_LOGSTD_INITIALIZER):
    self._policy_layers = policy_layers
    self._value_layers = value_layers
    self._action_size = action_size
    self._mean_weights_initializer = mean_weights_initializer
    self._logstd_initializer = logstd_initializer
    self._cell = tf.contrib.rnn.GRUBlockCell(100)

  @property
  def state_size(self):
    return self._cell.state_size

  @property
  def output_size(self):
    return (self._action_size, self._action_size, tf.TensorShape([]))

  def __call__(self, observation, state):
    with tf.variable_scope('policy'):
      x = tf.contrib.layers.flatten(observation)
      for size in self._policy_layers[:-1]:
        x = tf.contrib.layers.fully_connected(x, size, tf.nn.relu)
      x, state = self._cell(x, state)
      mean = tf.contrib.layers.fully_connected(x,
                                               self._action_size,
                                               tf.tanh,
                                               weights_initializer=self._mean_weights_initializer)
      logstd = tf.get_variable('logstd', mean.shape[1:], tf.float32, self._logstd_initializer)
      logstd = tf.tile(logstd[None, ...], [tf.shape(mean)[0]] + [1] * logstd.shape.ndims)
    with tf.variable_scope('value'):
      x = tf.contrib.layers.flatten(observation)
      for size in self._value_layers:
        x = tf.contrib.layers.fully_connected(x, size, tf.nn.relu)
      value = tf.contrib.layers.fully_connected(x, 1, None)[:, 0]
    return (mean, logstd, value), state



"""Trajectory Generator for in-place stepping motion for quadruped robot."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import numpy as np

TWO_PI = 2 * math.pi


def _get_actions_asymmetric_sine(phase, tg_params):
  """Returns the leg extension given current phase of TG and parameters.

  Args:
    phase: a number in [0, 2pi) representing current leg phase
    tg_params: a dictionary of tg parameters:
      stance_lift_cutoff -- switches the TG between stance (phase < cutoff) and
                            lift (phase > cutoff) phase
      amplitude_swing -- amplitude in swing phase
      amplitude_lift -- amplitude in lift phase
      center_extension -- center of leg extension
  """
  stance_lift_cutoff = tg_params['stance_lift_cutoff']
  a_prime = np.where(phase < stance_lift_cutoff, tg_params['amplitude_stance'],
                     tg_params['amplitude_lift'])
  scaled_phase = np.where(
      phase > stance_lift_cutoff, np.pi + (phase - stance_lift_cutoff) /
      (TWO_PI - stance_lift_cutoff) * np.pi, phase / stance_lift_cutoff * np.pi)
  return tg_params['center_extension'] + a_prime * np.sin(scaled_phase)


def step(current_phases, leg_frequencies, dt, tg_params):
  """Steps forward the in-place trajectory generator.

  Args:
    current_phases: phases of each leg.
    leg_frequencies: the frequency to proceed the phase of each leg.
    dt: amount of time (sec) between consecutive time steps.
    tg_params: a set of parameters for trajectory generator, see the docstring
      of "_get_actions_asymmetric_sine" for details.

  Returns:
    actions: leg swing/extensions as output by the trajectory generator.
    new_state: new swing/extension.
  """
  new_phases = np.fmod(current_phases + TWO_PI * leg_frequencies * dt, TWO_PI)
  extensions = []
  for leg_id in range(4):
    extensions.append(
        _get_actions_asymmetric_sine(new_phases[..., leg_id], tg_params))
  return new_phases, extensions


def reset():
  return np.array([0, np.pi * 0.5, np.pi, np.pi * 1.5])



# Lint as: python3
"""A sensor prototype class.

The concept is explained in: go/minitaur-gym-redesign-1.1
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from typing import Any, Iterable, Optional, Sequence, Text, Tuple, Union

import gin
import gym
import numpy as np

from pybullet_envs.minitaur.robots import robot_base
from pybullet_envs.minitaur.robots import time_ordered_buffer

_ARRAY = Sequence[float]
_FloatOrArray = Union[float, _ARRAY]
_DataTypeList = Iterable[Any]

# For sensor with multiput outputs, key of the main observation in output dict.
MAIN_OBS_KEY = ""

# This allows referencing np.float32 in gin config files. For example:
# lidar_sensor.LidarSensor.dtype = @np.float32
gin.external_configurable(np.float32, module="np")
gin.external_configurable(np.float64, module="np")
gin.external_configurable(np.uint8, module="np")


# Observation blenders take a pair of low/high values. The low/high is measured
# by the latency of the observation. So the low value is actually newer in time
# and high value older. The coeff [0, 1] can be thinked as the distance between
# the low and high value value, with 0 being 100% low value and 1 as 100% high
# value.
def linear_obs_blender(low_value: Any, high_value: Any, coeff: float):
  """Linear interpolation of low/high values based on coefficient value."""
  return low_value * (1 - coeff) + high_value * coeff


def closest_obs_blender(low_value: Any, high_value: Any, coeff: float):
  """Choosing the high or low value based on coefficient value."""
  return low_value if coeff < 0.5 else high_value


def newer_obs_blender(low_value: Any, unused_high_value: Any,
                      unused_coeff: float):
  """Always choosing low value, which is the newer value between low/high."""
  return low_value


def older_obs_blender(unused_low_value: Any, high_value: Any,
                      unused_coeff: float):
  """Always choosing the high value, which is the older value between low/high."""
  return high_value


@gin.configurable
class Sensor(object):
  """A prototype class of sensors."""

  def __init__(
      self,
      name: Text,
      sensor_latency: _FloatOrArray,
      interpolator_fn: Any,
      enable_debug_visualization: bool = False,
  ):
    """A basic constructor of the sensor.

    We do not provide a robot instance during __init__, as robot instances may
    be reloaded/recreated during the simulation.

    Args:
      name: the name of the sensor
      sensor_latency: There are two ways to use this expected sensor latency.
        For both methods, the latency should be in the same unit as the sensor
        data timestamp. 1. As a single float number, the observation will be a
        1D array. For real robots, this should be set to 0.0. 2. As an array of
        floats, the observation will be a 2D array based on how long the history
        need to be. Thus, [0.0, 0.1, 0.2] is a history length of 3. Observations
        are stacked on a new axis appended after existing axes.
      interpolator_fn: Function that controls how to interpolate the two values
        that is returned from the time ordered buffer.
      enable_debug_visualization: Whether to draw debugging visualization.
    """
    self._robot = None
    self._name = name
    # Observation space will be implemented by derived classes.
    self._observation_space = None
    self._sensor_latency = sensor_latency
    self._single_latency = True if isinstance(sensor_latency,
                                              (float, int)) else False
    self._enable_debug_visualization = enable_debug_visualization
    if not self._is_valid_latency():
      raise ValueError("sensor_latency is expected to be a non-negative number "
                       "or a non-empty list of non-negative numbers.")
    self._interpolator_fn = interpolator_fn or newer_obs_blender
    self._axis = -1
    timespan = sensor_latency if self._single_latency else max(sensor_latency)
    self._observation_buffer = time_ordered_buffer.TimeOrderedBuffer(
        max_buffer_timespan=timespan)

  def _is_valid_latency(self):
    if self._single_latency:
      return self._sensor_latency >= 0
    if self._sensor_latency:
      return all(value >= 0 for value in self._sensor_latency)
    return False

  def get_name(self) -> Text:
    return self._name

  @property
  def is_single_latency(self) -> bool:
    return self._single_latency

  @property
  def observation_space(self) -> gym.spaces.Space:
    return self._observation_space

  @property
  def enable_debug_visualization(self):
    return self._enable_debug_visualization

  @enable_debug_visualization.setter
  def enable_debug_visualization(self, enable):
    self._enable_debug_visualization = enable

  def get_observation_datatype(self):
    """Returns the data type for the numpy structured array.

    It is recommended to define a list of tuples: (name, datatype, shape)
    Reference: https://docs.scipy.org/doc/numpy-1.15.0/user/basics.rec.html
    Ex:
      return [('motor_angles', np.float64, (8, ))]  # motor angle sensor
      return [('IMU_x', np.float64), ('IMU_z', np.float64), ] # IMU
    Will be deprecated (b/150818246) in favor of observation_space.

    Returns:
      datatype: a list of data types.
    """
    raise NotImplementedError("Deprecated. Are you using the old robot class?")

  def get_lower_bound(self):
    """Returns the lower bound of the observation.

    Will be deprecated (b/150818246) in favor of observation_space.

    Returns:
      lower_bound: the lower bound of sensor values in np.array format
    """
    raise NotImplementedError("Deprecated. Are you using the old robot class?")

  def get_upper_bound(self):
    """Returns the upper bound of the observation.

    Will be deprecated (b/150818246) in favor of observation_space.

    Returns:
      upper_bound: the upper bound of sensor values in np.array format
    """
    raise NotImplementedError("Deprecated. Are you using the old robot class?")

  def _get_original_observation(self) -> Tuple[float, Any]:
    """Gets the non-modified observation.

    Different from the get_observation, which can pollute and sensor data with
    noise and latency, this method shall return the best effort measurements of
    the sensor. For simulated robots, this will return the clean data. For reals
    robots, just return the measurements as is. All inherited class shall
    implement this method.

    Returns:
      The timestamp and the original sensor measurements.

    Raises:
      NotImplementedError for the base class.

    """
    raise NotImplementedError("Not implemented for base class." "")

  def get_observation(self):
    """Returns the observation data.

    Returns:
      observation: the observed sensor values in np.array format
    """
    obs = self._observation_buffer.get_delayed_value(self._sensor_latency)

    if self._single_latency:
      if isinstance(self._observation_space, gym.spaces.Dict):
        return self._interpolator_fn(obs.value_0, obs.value_1, obs.coeff)
      else:
        return np.asarray(
            self._interpolator_fn(obs.value_0, obs.value_1, obs.coeff))
    else:
      if isinstance(self._observation_space, gym.spaces.Dict):
        # interpolate individual sub observation
        interpolated = [
            self._interpolator_fn(data.value_0, data.value_1, data.coeff)
            for data in obs
        ]

        stacked_per_sub_obs = {}
        for k in interpolated[0]:
          stacked_per_sub_obs[k] = np.stack(
              np.asarray([d[k] for d in interpolated]), axis=self._axis)
        return stacked_per_sub_obs
      else:
        obs = np.asarray([
            self._interpolator_fn(data.value_0, data.value_1, data.coeff)
            for data in obs
        ])
        return np.stack(obs, axis=self._axis)

  def set_robot(self, robot: robot_base.RobotBase):
    """Set a robot instance."""
    self._robot = robot

  def get_robot(self):
    """Returns the robot instance."""
    return self._robot

  def on_reset(self, env):
    """A callback function for the reset event.

    Args:
      env: the environment who invokes this callback function.
    """
    self._env = env
    self._observation_buffer.reset()
    self.on_new_observation()

  def on_step(self, env):
    """A callback function for the control step event.

    Args:
      env: the environment who invokes this callback function.
    """
    pass

  def visualize(self):
    """Visualizes the sensor information."""
    pass

  def on_new_observation(self):
    """A callback for each observation received.

    To be differentiated from on_step, which will be called only once per
    control step (i.e. env.step), this API will be called everytime in the
    substep/action repeat loop, when new observations are expected. Each derived
    sensor class should implement this API by implementing:

    my_obs  = call env/robot api to get the observation
    self._observation_buffer.add(my_obs)
    """
    timestamp, obs = self._get_original_observation()
    if self._enable_debug_visualization:
      self.visualize()
    self._observation_buffer.add(timestamp, obs)

  def on_terminate(self, env):
    """A callback function for the terminate event.

    Args:
      env: the environment who invokes this callback function.
    """
    pass

  def _stack_space(self,
                   space: Union[gym.spaces.Box, gym.spaces.Dict],
                   dtype: np.dtype = None) -> Any:
    """Returns stacked version of observation space.

    This stacks a gym.spaces.Box or gym.spaces.Dict action space based on the
    length of the sensor latency and the axis for stacking specified in the
    sensor. A gym.spaces.Box is just stacked, but a gym.spaces.Dict is
    recursively stacked, preserving its dictionary structure while stacking
    any gym.spaces.Box contained within. For example, the input action space:

        gym.spaces.Dict({
            'space_1': gym.spaces.Box(low=0, high=10, shape=(1,)),
            'space_2': gym.spaces.Dict({
                'space_3': gym.spaces.Box(low=0, high=10, shape=(2,)),
                }),
        }))

    would be converted to the following if sensor latency was [0, 1]:

        gym.spaces.Dict({
            'space_1': gym.spaces.Box(low=0, high=10, shape=(1, 2)),
            'space_2': gym.spaces.Dict({
                'space_3': gym.spaces.Box(low=0, high=10, shape=(2, 2)),
                }),
        }))

    Args:
      space: A gym.spaces.Dict or gym.spaces.Box to be stacked.
      dtype: Datatype for the stacking.

    Returns:
      stacked_space: A stacked version of the action space.
    """
    if self._single_latency:
      return space

    # Allow sensors such as last_action_sensor to override the dtype.
    dtype = dtype or space.dtype

    if isinstance(space, gym.spaces.Box):
      return self._stack_space_box(space, dtype)
    elif isinstance(space, gym.spaces.Dict):
      return self._stack_space_dict(space, dtype)
    else:
      raise ValueError(f"Space {space} is an unsupported type.")

  def _stack_space_box(self, space: gym.spaces.Box,
                       dtype: np.dtype) -> gym.spaces.Box:
    """Returns stacked version of a box observation space.

    This stacks a gym.spaces.Box action space based on the length of the sensor
    latency and the axis for stacking specified in the sensor.

    Args:
      space: A gym.spaces.Box to be stacked.
      dtype: Datatype for the stacking

    Returns:
      stacked_space: A stacked version of the gym.spaces.Box action space.
    """
    length = len(self._sensor_latency)
    stacked_space = gym.spaces.Box(
        low=np.repeat(
            np.expand_dims(space.low, axis=self._axis), length,
            axis=self._axis),
        high=np.repeat(
            np.expand_dims(space.high, axis=self._axis),
            length,
            axis=self._axis),
        dtype=dtype)

    return stacked_space

  def _stack_space_dict(self, space: gym.spaces.Dict,
                        dtype: np.dtype) -> gym.spaces.Dict:
    """Returns stacked version of a dict observation space.

    This stacks a gym.spaces.Dict action space based on the length of the sensor
    latency and the recursive structure of the gym.spaces.Dict itself.

    Args:
      space: A gym.spaces.Dict to be stacked.
      dtype: Datatype for the stacking.

    Returns:
      stacked_space: A stacked version of the dictionary action space.
    """
    return gym.spaces.Dict([
        (k, self._stack_space(v, dtype)) for k, v in space.spaces.items()
    ])

  def _encode_obs_dict_keys(self, obs_dict):
    """Encodes sub obs keys of observation dict or observsation space dict."""
    return {encode_sub_obs_key(self, k): v for k, v in obs_dict.items()}


class BoxSpaceSensor(Sensor):
  """A prototype class of sensors with Box shapes."""

  def __init__(self,
               name: Text,
               shape: Tuple[int, ...],
               lower_bound: _FloatOrArray = -np.pi,
               upper_bound: _FloatOrArray = np.pi,
               dtype=np.float64) -> None:
    """Constructs a box type sensor.

    Will be deprecated (b/150818246) once we switch to gym spaces.
    Args:
      name: the name of the sensor
      shape: the shape of the sensor values
      lower_bound: the lower_bound of sensor value, in float or np.array.
      upper_bound: the upper_bound of sensor value, in float or np.array.
      dtype: data type of sensor value
    """
    super(BoxSpaceSensor, self).__init__(
        name=name, sensor_latency=0.0, interpolator_fn=newer_obs_blender)
    self._shape = shape
    self._dtype = dtype

    if isinstance(lower_bound, float):
      self._lower_bound = np.full(shape, lower_bound, dtype=dtype)
    else:
      self._lower_bound = np.array(lower_bound)

    if isinstance(upper_bound, float):
      self._upper_bound = np.full(shape, upper_bound, dtype=dtype)
    else:
      self._upper_bound = np.array(upper_bound)

  def set_robot(self, robot):
    # Since all old robot class do not inherit from RobotBase, we can enforce
    # the checking here.
    if isinstance(robot, robot_base.RobotBase):
      raise ValueError(
          "Cannot use new robot interface RobotBase with old sensor calss.")
    self._robot = robot

  def get_shape(self) -> Tuple[int, ...]:
    return self._shape

  def get_dimension(self) -> int:
    return len(self._shape)

  def get_dtype(self):
    return self._dtype

  def get_observation_datatype(self) -> _DataTypeList:
    """Returns box-shape data type."""
    return [(self._name, self._dtype, self._shape)]

  def get_lower_bound(self) -> _ARRAY:
    """Returns the computed lower bound."""
    return self._lower_bound

  def get_upper_bound(self) -> _ARRAY:
    """Returns the computed upper bound."""
    return self._upper_bound

  def get_observation(self) -> np.ndarray:
    return np.asarray(self._get_observation(), dtype=self._dtype)

  def _get_original_observation(self) -> Tuple[float, Any]:
    # Maintains compatibility with the new sensor classes."""
    raise NotImplementedError("Not implemented for this class.")

  def on_new_observation(self):
    # Maintains compatibility with the new sensor classes."""
    pass


def encode_sub_obs_key(s: Sensor, sub_obs_name: Optional[Text]):
  """Returns a sub observation key for use in observation dictionary."""
  if sub_obs_name == MAIN_OBS_KEY:
    return s.get_name()
  else:
    return f"{s.get_name()}/{sub_obs_name}"



"""An example to run of the minitaur gym environment with standing up goal.

"""

import math

import numpy as np
import tf.compat.v1 as tf

from pybullet_envs.minitaur.envs import minitaur_stand_gym_env


def StandUpExample():
  """An example that the minitaur stands up."""
  steps = 1000
  environment = minitaur_stand_gym_env.MinitaurStandGymEnv(render=True,
                                                           motor_velocity_limit=np.inf)
  action = [0.5]
  _, _, done, _ = environment.step(action)
  for t in range(steps):
    # A policy that oscillates between -1 and 1
    action = [math.sin(t * math.pi * 0.01)]
    _, _, done, _ = environment.step(action)
    if done:
      break


def main(unused_argv):
  StandUpExample()


if __name__ == "__main__":
  tf.logging.set_verbosity(tf.logging.INFO)
  main("unused")



r"""Running a pre-trained ppo agent on minitaur_reactive_env."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import time

import inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(os.path.dirname(currentdir)))
print("parentdir=", parentdir)
os.sys.path.insert(0, parentdir)

import tf.compat.v1 as tf
from pybullet_envs.minitaur.agents.scripts import utility
import pybullet_data
from pybullet_envs.minitaur.envs import simple_ppo_agent

flags = tf.app.flags
FLAGS = tf.app.flags.FLAGS
LOG_DIR = os.path.join(pybullet_data.getDataPath(), "policies/ppo/minitaur_reactive_env")
CHECKPOINT = "model.ckpt-14000000"


def main(argv):
  del argv  # Unused.
  config = utility.load_config(LOG_DIR)
  policy_layers = config.policy_layers
  value_layers = config.value_layers
  env = config.env(render=True)
  network = config.network

  with tf.Session() as sess:
    agent = simple_ppo_agent.SimplePPOPolicy(sess,
                                             env,
                                             network,
                                             policy_layers=policy_layers,
                                             value_layers=value_layers,
                                             checkpoint=os.path.join(LOG_DIR, CHECKPOINT))

    sum_reward = 0
    observation = env.reset()
    while True:
      action = agent.get_action([observation])
      observation, reward, done, _ = env.step(action[0])
      time.sleep(0.002)
      sum_reward += reward
      if done:
        break
    tf.logging.info("reward: %s", sum_reward)


if __name__ == "__main__":
  tf.app.run(main)



import pandas as pd
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

content = []

with open("output/agent0_log.txt", "r") as f:
    content = f.readlines()

for i in range(len(content)):
    content[i] = ' '.join(content[i].split())
    content[i] = content[i].split(" ")

header = content[0]
df = pd.DataFrame(content[1:], columns=header, dtype=float)

tr_vals = df["Train_Return"].values
te_vals = df["Test_Return"].values
x = range(0, len(tr_vals)*400, 400)

fig, ax = plt.subplots()
ax.plot(x, tr_vals, label="Train")
ax.plot(x, te_vals, label="Test")
ax.set_xlim(0, (len(tr_vals)+1)*400)
ax.set_xlabel('Iterations')
ax.set_ylim(0, 100, 10)
ax.set_ylabel('Return')
ax.legend()
plt.savefig("output/rewards_log.png")



# Lint as: python3
"""A generic PD motor model."""

from typing import Tuple, Union
import gin
import numpy as np

from pybullet_envs.minitaur.robots import robot_config
from pybullet_envs.minitaur.robots import time_ordered_buffer

_DEFAULT_BUFFER_SIZE = 200

_HYBRID_ACTION_LEN = len(robot_config.HybridActionIndex)
_HYBRID_POS_INDEX = robot_config.HybridActionIndex.POSITION.value
_HYBRID_KP_INDEX = robot_config.HybridActionIndex.POSITION_GAIN.value
_HYBRID_VEL_INDEX = robot_config.HybridActionIndex.VELOCITY.value
_HYBRID_KD_INDEX = robot_config.HybridActionIndex.VELOCITY_GAIN.value
_HYBRID_TORQUE_INDEX = robot_config.HybridActionIndex.TORQUE.value


def _convert_to_np_array(inputs: Union[float, Tuple[float], np.ndarray], dim):
  """Converts the inputs to a numpy array.

  Args:
    inputs: The input scalar or array.
    dim: The dimension of the converted numpy array.

  Returns:
    The converted numpy array.

  Raises:
    ValueError: If the inputs is an array whose dimension does not match the
    provied dimension.
  """
  outputs = None
  if isinstance(inputs, (tuple, np.ndarray)):
    outputs = np.array(inputs)
  else:
    outputs = np.full(dim, inputs)

  if len(outputs) != dim:
    raise ValueError("The inputs array has a different dimension {}"
                     " than provided, which is {}.".format(len(outputs), dim))

  return outputs


@gin.configurable
class HybridMotorModel(object):
  """A simple motor model that supports proportional and derivative control.

    When in POSITION mode, the torque is calculated according to the difference
    between current and desired joint angle, as well as the joint velocity
    differences. For more information about PD control, please refer to:
    https://en.wikipedia.org/wiki/PID_controller.

    The model supports a HYBRID mode in which each motor command can be a tuple
    (desired_motor_angle, position_gain, desired_motor_velocity, velocity_gain,
    torque).
  """

  def __init__(
      self,
      num_motors: int,
      pd_latency: float = 0,
      motor_control_mode=robot_config.MotorControlMode.POSITION,
      kp: Union[float, Tuple[float], np.ndarray] = 60,
      kd: Union[float, Tuple[float], np.ndarray] = 1,
      strength_ratios: Union[float, Tuple[float], np.ndarray] = 1,
      torque_lower_limits: Union[float, Tuple[float], np.ndarray] = None,
      torque_upper_limits: Union[float, Tuple[float], np.ndarray] = None,
  ):
    """Initializes the class.

    Args:
      num_motors: The number of motors for parallel computation.
      pd_latency: Simulates the motor controller's latency in reading motor
        angles and velocities.
      motor_control_mode: Can be POSITION, TORQUE, or HYBRID. In POSITION
        control mode, the PD formula is used to track a desired position and a
        zero desired velocity. In TORQUE control mode, we assume a pass through
        of the provided torques. In HYBRID control mode, the users need to
        provie (desired_position, position_gain, desired_velocity,
        velocity_gain, feedfoward_torque) for each motor.
      kp: The default position gains for motors.
      kd: The default velocity gains for motors.
      strength_ratios: The scaling ratio for motor torque outputs. This can be
        useful for quick debugging when sim-to-real gap is observed in the
        actuator behavior.
      torque_lower_limits: The lower bounds for torque outputs.
      torque_upper_limits: The upper bounds for torque outputs. The output
        torques will be clipped by the lower and upper bounds.

    Raises:
      ValueError: If the number of motors provided is negative or zero.
    """
    if num_motors <= 0:
      raise ValueError(
          "Number of motors must be positive, not {}".format(num_motors))
    self._num_motors = num_motors
    self._zero_array = np.full(num_motors, 0)
    self._pd_latency = pd_latency
    self._hybrid_command_dim = _HYBRID_ACTION_LEN * self._num_motors
    self.set_motor_gains(kp, kd)
    self.set_strength_ratios(strength_ratios)
    self._torque_lower_limits = None
    if torque_lower_limits:
      self._torque_lower_limits = _convert_to_np_array(torque_lower_limits,
                                                       self._num_motors)

    self._torque_upper_limits = None
    if torque_upper_limits:
      self._torque_upper_limits = _convert_to_np_array(torque_upper_limits,
                                                       self._num_motors)
    self._motor_control_mode = motor_control_mode

    # The history buffer is used to simulate the pd latency effect.
    # TODO(b/157786642): remove hacks on duplicate timestep once the sim clock
    # is fixed.
    self._observation_buffer = time_ordered_buffer.TimeOrderedBuffer(
        max_buffer_timespan=pd_latency,
        error_on_duplicate_timestamp=False,
        replace_value_on_duplicate_timestamp=True)

  def set_strength_ratios(
      self,
      strength_ratios: Union[float, Tuple[float], np.ndarray],
  ):
    """Sets the strength of each motor relative to the default value.

    Args:
      strength_ratios: The relative strength of motor output, ranging from [0,
        1] inclusive.
    """
    self._strength_ratios = np.clip(
        _convert_to_np_array(strength_ratios, self._num_motors), 0, 1)

  def set_motor_gains(
      self,
      kp: Union[float, Tuple[float], np.ndarray],
      kd: Union[float, Tuple[float], np.ndarray],
  ):
    """Sets the gains of all motors.

    These gains are PD gains for motor positional control. kp is the
    proportional gain and kd is the derivative gain.

    Args:
      kp: Proportional gain of the motors.
      kd: Derivative gain of the motors.
    """
    self._kp = _convert_to_np_array(kp, self._num_motors)
    self._kd = _convert_to_np_array(kd, self._num_motors)

  def get_motor_gains(self):
    """Get the PD gains of all motors.

    Returns:
      Proportional and derivative gain of the motors.
    """
    return self._kp, self._kd

  def reset(self):
    self._observation_buffer.reset()

  def update(self, timestamp, true_motor_positions: np.ndarray,
             true_motor_velocities: np.ndarray):
    # Push these to the buffer
    self._observation_buffer.add(timestamp,
                                 (true_motor_positions, true_motor_velocities))

  def get_motor_torques(
      self,
      motor_commands: np.ndarray,
      motor_control_mode=None) -> Tuple[np.ndarray, np.ndarray]:
    """Computes the motor torques.

    Args:
      motor_commands: The desired motor angle if the motor is in position
        control mode. The pwm signal if the motor is in torque control mode.
      motor_control_mode: A MotorControlMode enum.

    Returns:
      observed_torque: The torque observed. This emulates the limitations in
      torque measurement, which is generally obtained from current estimations.
      actual_torque: The torque that needs to be applied to the motor.

    Raises:
      NotImplementedError if the motor_control_mode is not supported.

    """
    if not motor_control_mode:
      motor_control_mode = self._motor_control_mode

    motor_torques = None

    if motor_control_mode is robot_config.MotorControlMode.TORQUE:
      motor_torques = motor_commands

    if motor_control_mode is robot_config.MotorControlMode.POSITION:
      motor_torques = self._compute_pd_torques(
          desired_motor_angles=motor_commands,
          kp=self._kp,
          desired_motor_velocities=self._zero_array,
          kd=self._kd)

    if motor_control_mode is robot_config.MotorControlMode.HYBRID:
      motor_torques = self._compute_hybrid_action_torques(motor_commands)

    if motor_torques is None:
      raise ValueError(
          "{} is not a supported motor control mode".format(motor_control_mode))

    # Rescale and clip the motor torques as needed.
    motor_torques = self._strength_ratios * motor_torques
    if (self._torque_lower_limits is not None or
        self._torque_upper_limits is not None):
      motor_torques = np.clip(motor_torques, self._torque_lower_limits,
                              self._torque_upper_limits)

    return motor_torques, motor_torques

  def get_motor_states(self, latency=None):
    """Computes observation of motor angle and velocity under latency."""
    if latency is None:
      latency = self._pd_latency
    buffer = self._observation_buffer.get_delayed_value(latency)
    angle_vel_t0 = buffer.value_0
    angle_vel_t1 = buffer.value_1
    coeff = buffer.coeff

    pos_idx = 0
    motor_angles = angle_vel_t0[pos_idx] * (
        1 - coeff) + coeff * angle_vel_t1[pos_idx]
    vel_idx = 1
    motor_velocities = angle_vel_t0[vel_idx] * (
        1 - coeff) + coeff * angle_vel_t1[vel_idx]
    return motor_angles, motor_velocities

  def _compute_pd_torques(
      self,
      desired_motor_angles: np.ndarray,
      kp: np.ndarray,
      desired_motor_velocities,
      kd: np.ndarray,
  ) -> Tuple[np.ndarray, np.ndarray]:
    """Computes the pd torques.

    Args:
      desired_motor_angles: The motor angles to track.
      kp: The position gains.
      desired_motor_velocities: The motor velocities to track.
      kd: The velocity gains.

    Returns:
      The computed motor torques.
    """
    motor_angles, motor_velocities = self.get_motor_states()
    motor_torques = -kp * (motor_angles - desired_motor_angles) - kd * (
        motor_velocities - desired_motor_velocities)

    return motor_torques

  def _compute_hybrid_action_torques(
      self, motor_commands: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Computes the pd torques in the HYBRID mode."""
    assert len(motor_commands) == self._hybrid_command_dim
    kp = motor_commands[_HYBRID_KP_INDEX::_HYBRID_ACTION_LEN]
    kd = motor_commands[_HYBRID_KD_INDEX::_HYBRID_ACTION_LEN]
    desired_motor_angles = motor_commands[_HYBRID_POS_INDEX::_HYBRID_ACTION_LEN]
    desired_motor_velocities = motor_commands[
        _HYBRID_VEL_INDEX::_HYBRID_ACTION_LEN]
    additional_torques = motor_commands[
        _HYBRID_TORQUE_INDEX::_HYBRID_ACTION_LEN]

    return self._compute_pd_torques(desired_motor_angles, kp,
                                    desired_motor_velocities,
                                    kd) + additional_torques



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: world_asset.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='world_asset.proto',
  package='robotics.messages',
  syntax='proto3',
  serialized_options=b'\370\001\001',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x11world_asset.proto\x12\x11robotics.messages\":\n\x0cQQuaternionf\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02\x12\t\n\x01w\x18\x04 \x01(\x02\",\n\tVVector3f\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02\"k\n\x06Labels\x12\x33\n\x05label\x18\x01 \x03(\x0b\x32$.robotics.messages.Labels.LabelEntry\x1a,\n\nLabelEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xa4\x01\n\x0cGeneralBox3f\x12,\n\x06\x63\x65nter\x18\x01 \x01(\x0b\x32\x1c.robotics.messages.VVector3f\x12\x30\n\ndimensions\x18\x02 \x01(\x0b\x32\x1c.robotics.messages.VVector3f\x12\x34\n\x0borientation\x18\x03 \x01(\x0b\x32\x1f.robotics.messages.QQuaternionf\"m\n\x06Object\x12\n\n\x02id\x18\x01 \x01(\t\x12\r\n\x05label\x18\x02 \x01(\t\x12\x35\n\x0c\x62ounding_box\x18\x03 \x01(\x0b\x32\x1f.robotics.messages.GeneralBox3f\x12\x11\n\tmesh_path\x18\x04 \x01(\t\"9\n\tPolygon3f\x12,\n\x06vertex\x18\x01 \x03(\x0b\x32\x1c.robotics.messages.VVector3f\"\xb2\x01\n\x06Region\x12\n\n\x02id\x18\x01 \x01(\t\x12\r\n\x05label\x18\x02 \x01(\t\x12\x37\n\x0c\x62ounding_box\x18\x03 \x01(\x0b\x32\x1f.robotics.messages.GeneralBox3fH\x00\x12/\n\x07polygon\x18\x05 \x01(\x0b\x32\x1c.robotics.messages.Polygon3fH\x00\x12\x11\n\tmesh_path\x18\x06 \x01(\tB\x10\n\x0espatial_extent\"\xfb\x01\n\nWorldAsset\x12\x30\n\robject_labels\x18\x01 \x01(\x0b\x32\x19.robotics.messages.Labels\x12*\n\x07objects\x18\x02 \x03(\x0b\x32\x19.robotics.messages.Object\x12\x30\n\rregion_labels\x18\x03 \x01(\x0b\x32\x19.robotics.messages.Labels\x12*\n\x07regions\x18\x04 \x03(\x0b\x32\x19.robotics.messages.Region\x12\x11\n\tmesh_path\x18\x05 \x01(\t\x12\x1e\n\x16segmentation_mesh_path\x18\x06 \x01(\tB\x03\xf8\x01\x01\x62\x06proto3'
)




_QQUATERNIONF = _descriptor.Descriptor(
  name='QQuaternionf',
  full_name='robotics.messages.QQuaternionf',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='robotics.messages.QQuaternionf.x', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='y', full_name='robotics.messages.QQuaternionf.y', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='z', full_name='robotics.messages.QQuaternionf.z', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='w', full_name='robotics.messages.QQuaternionf.w', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=40,
  serialized_end=98,
)


_VVECTOR3F = _descriptor.Descriptor(
  name='VVector3f',
  full_name='robotics.messages.VVector3f',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='robotics.messages.VVector3f.x', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='y', full_name='robotics.messages.VVector3f.y', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='z', full_name='robotics.messages.VVector3f.z', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=100,
  serialized_end=144,
)


_LABELS_LABELENTRY = _descriptor.Descriptor(
  name='LabelEntry',
  full_name='robotics.messages.Labels.LabelEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='robotics.messages.Labels.LabelEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='robotics.messages.Labels.LabelEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=209,
  serialized_end=253,
)

_LABELS = _descriptor.Descriptor(
  name='Labels',
  full_name='robotics.messages.Labels',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='label', full_name='robotics.messages.Labels.label', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_LABELS_LABELENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=146,
  serialized_end=253,
)


_GENERALBOX3F = _descriptor.Descriptor(
  name='GeneralBox3f',
  full_name='robotics.messages.GeneralBox3f',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='center', full_name='robotics.messages.GeneralBox3f.center', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dimensions', full_name='robotics.messages.GeneralBox3f.dimensions', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='orientation', full_name='robotics.messages.GeneralBox3f.orientation', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=256,
  serialized_end=420,
)


_OBJECT = _descriptor.Descriptor(
  name='Object',
  full_name='robotics.messages.Object',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='robotics.messages.Object.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='label', full_name='robotics.messages.Object.label', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bounding_box', full_name='robotics.messages.Object.bounding_box', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mesh_path', full_name='robotics.messages.Object.mesh_path', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=422,
  serialized_end=531,
)


_POLYGON3F = _descriptor.Descriptor(
  name='Polygon3f',
  full_name='robotics.messages.Polygon3f',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='vertex', full_name='robotics.messages.Polygon3f.vertex', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=533,
  serialized_end=590,
)


_REGION = _descriptor.Descriptor(
  name='Region',
  full_name='robotics.messages.Region',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='robotics.messages.Region.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='label', full_name='robotics.messages.Region.label', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bounding_box', full_name='robotics.messages.Region.bounding_box', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='polygon', full_name='robotics.messages.Region.polygon', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mesh_path', full_name='robotics.messages.Region.mesh_path', index=4,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='spatial_extent', full_name='robotics.messages.Region.spatial_extent',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=593,
  serialized_end=771,
)


_WORLDASSET = _descriptor.Descriptor(
  name='WorldAsset',
  full_name='robotics.messages.WorldAsset',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='object_labels', full_name='robotics.messages.WorldAsset.object_labels', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='objects', full_name='robotics.messages.WorldAsset.objects', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='region_labels', full_name='robotics.messages.WorldAsset.region_labels', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='regions', full_name='robotics.messages.WorldAsset.regions', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mesh_path', full_name='robotics.messages.WorldAsset.mesh_path', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='segmentation_mesh_path', full_name='robotics.messages.WorldAsset.segmentation_mesh_path', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=774,
  serialized_end=1025,
)

_LABELS_LABELENTRY.containing_type = _LABELS
_LABELS.fields_by_name['label'].message_type = _LABELS_LABELENTRY
_GENERALBOX3F.fields_by_name['center'].message_type = _VVECTOR3F
_GENERALBOX3F.fields_by_name['dimensions'].message_type = _VVECTOR3F
_GENERALBOX3F.fields_by_name['orientation'].message_type = _QQUATERNIONF
_OBJECT.fields_by_name['bounding_box'].message_type = _GENERALBOX3F
_POLYGON3F.fields_by_name['vertex'].message_type = _VVECTOR3F
_REGION.fields_by_name['bounding_box'].message_type = _GENERALBOX3F
_REGION.fields_by_name['polygon'].message_type = _POLYGON3F
_REGION.oneofs_by_name['spatial_extent'].fields.append(
  _REGION.fields_by_name['bounding_box'])
_REGION.fields_by_name['bounding_box'].containing_oneof = _REGION.oneofs_by_name['spatial_extent']
_REGION.oneofs_by_name['spatial_extent'].fields.append(
  _REGION.fields_by_name['polygon'])
_REGION.fields_by_name['polygon'].containing_oneof = _REGION.oneofs_by_name['spatial_extent']
_WORLDASSET.fields_by_name['object_labels'].message_type = _LABELS
_WORLDASSET.fields_by_name['objects'].message_type = _OBJECT
_WORLDASSET.fields_by_name['region_labels'].message_type = _LABELS
_WORLDASSET.fields_by_name['regions'].message_type = _REGION
DESCRIPTOR.message_types_by_name['QQuaternionf'] = _QQUATERNIONF
DESCRIPTOR.message_types_by_name['VVector3f'] = _VVECTOR3F
DESCRIPTOR.message_types_by_name['Labels'] = _LABELS
DESCRIPTOR.message_types_by_name['GeneralBox3f'] = _GENERALBOX3F
DESCRIPTOR.message_types_by_name['Object'] = _OBJECT
DESCRIPTOR.message_types_by_name['Polygon3f'] = _POLYGON3F
DESCRIPTOR.message_types_by_name['Region'] = _REGION
DESCRIPTOR.message_types_by_name['WorldAsset'] = _WORLDASSET
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

QQuaternionf = _reflection.GeneratedProtocolMessageType('QQuaternionf', (_message.Message,), {
  'DESCRIPTOR' : _QQUATERNIONF,
  '__module__' : 'world_asset_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.QQuaternionf)
  })
_sym_db.RegisterMessage(QQuaternionf)

VVector3f = _reflection.GeneratedProtocolMessageType('VVector3f', (_message.Message,), {
  'DESCRIPTOR' : _VVECTOR3F,
  '__module__' : 'world_asset_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.VVector3f)
  })
_sym_db.RegisterMessage(VVector3f)

Labels = _reflection.GeneratedProtocolMessageType('Labels', (_message.Message,), {

  'LabelEntry' : _reflection.GeneratedProtocolMessageType('LabelEntry', (_message.Message,), {
    'DESCRIPTOR' : _LABELS_LABELENTRY,
    '__module__' : 'world_asset_pb2'
    # @@protoc_insertion_point(class_scope:robotics.messages.Labels.LabelEntry)
    })
  ,
  'DESCRIPTOR' : _LABELS,
  '__module__' : 'world_asset_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Labels)
  })
_sym_db.RegisterMessage(Labels)
_sym_db.RegisterMessage(Labels.LabelEntry)

GeneralBox3f = _reflection.GeneratedProtocolMessageType('GeneralBox3f', (_message.Message,), {
  'DESCRIPTOR' : _GENERALBOX3F,
  '__module__' : 'world_asset_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.GeneralBox3f)
  })
_sym_db.RegisterMessage(GeneralBox3f)

Object = _reflection.GeneratedProtocolMessageType('Object', (_message.Message,), {
  'DESCRIPTOR' : _OBJECT,
  '__module__' : 'world_asset_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Object)
  })
_sym_db.RegisterMessage(Object)

Polygon3f = _reflection.GeneratedProtocolMessageType('Polygon3f', (_message.Message,), {
  'DESCRIPTOR' : _POLYGON3F,
  '__module__' : 'world_asset_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Polygon3f)
  })
_sym_db.RegisterMessage(Polygon3f)

Region = _reflection.GeneratedProtocolMessageType('Region', (_message.Message,), {
  'DESCRIPTOR' : _REGION,
  '__module__' : 'world_asset_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.Region)
  })
_sym_db.RegisterMessage(Region)

WorldAsset = _reflection.GeneratedProtocolMessageType('WorldAsset', (_message.Message,), {
  'DESCRIPTOR' : _WORLDASSET,
  '__module__' : 'world_asset_pb2'
  # @@protoc_insertion_point(class_scope:robotics.messages.WorldAsset)
  })
_sym_db.RegisterMessage(WorldAsset)


DESCRIPTOR._options = None
_LABELS_LABELENTRY._options = None
# @@protoc_insertion_point(module_scope)



import numpy as np


def compute_return(rewards, gamma, td_lambda, val_t):
  # computes td-lambda return of path
  path_len = len(rewards)
  assert len(val_t) == path_len + 1

  return_t = np.zeros(path_len)
  last_val = rewards[-1] + gamma * val_t[-1]
  return_t[-1] = last_val

  for i in reversed(range(0, path_len - 1)):
    curr_r = rewards[i]
    next_ret = return_t[i + 1]
    curr_val = curr_r + gamma * ((1.0 - td_lambda) * val_t[i + 1] + td_lambda * next_ret)
    return_t[i] = curr_val

  return return_t



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Proximal Policy Optimization algorithm.

Based on John Schulman's implementation in Python and Theano:
https://github.com/joschu/modular_rl/blob/master/modular_rl/ppo.py
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools

try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf

from . import memory
from . import normalize
from . import utility


class PPOAlgorithm(object):
  """A vectorized implementation of the PPO algorithm by John Schulman."""

  def __init__(self, batch_env, step, is_training, should_log, config):
    """Create an instance of the PPO algorithm.

    Args:
      batch_env: In-graph batch environment.
      step: Integer tensor holding the current training step.
      is_training: Boolean tensor for whether the algorithm should train.
      should_log: Boolean tensor for whether summaries should be returned.
      config: Object containing the agent configuration as attributes.
    """
    self._batch_env = batch_env
    self._step = step
    self._is_training = is_training
    self._should_log = should_log
    self._config = config
    self._observ_filter = normalize.StreamingNormalize(self._batch_env.observ[0],
                                                       center=True,
                                                       scale=True,
                                                       clip=5,
                                                       name='normalize_observ')
    self._reward_filter = normalize.StreamingNormalize(self._batch_env.reward[0],
                                                       center=False,
                                                       scale=True,
                                                       clip=10,
                                                       name='normalize_reward')
    # Memory stores tuple of observ, action, mean, logstd, reward.
    template = (self._batch_env.observ[0], self._batch_env.action[0], self._batch_env.action[0],
                self._batch_env.action[0], self._batch_env.reward[0])
    self._memory = memory.EpisodeMemory(template, config.update_every, config.max_length, 'memory')
    self._memory_index = tf.Variable(0, False)
    use_gpu = self._config.use_gpu and utility.available_gpus()
    with tf.device('/gpu:0' if use_gpu else '/cpu:0'):
      # Create network variables for later calls to reuse.
      action_size = self._batch_env.action.shape[1].value
      self._network = tf.make_template('network',
                                       functools.partial(config.network, config, action_size))
      output = self._network(
          tf.zeros_like(self._batch_env.observ)[:, None], tf.ones(len(self._batch_env)))
      with tf.variable_scope('ppo_temporary'):
        self._episodes = memory.EpisodeMemory(template, len(batch_env), config.max_length,
                                              'episodes')
        if output.state is None:
          self._last_state = None
        else:
          # Ensure the batch dimension is set.
          tf.contrib.framework.nest.map_structure(
              lambda x: x.set_shape([len(batch_env)] + x.shape.as_list()[1:]), output.state)
          # pylint: disable=undefined-variable
          self._last_state = tf.contrib.framework.nest.map_structure(
              lambda x: tf.Variable(lambda: tf.zeros_like(x), False), output.state)
        self._last_action = tf.Variable(tf.zeros_like(self._batch_env.action),
                                        False,
                                        name='last_action')
        self._last_mean = tf.Variable(tf.zeros_like(self._batch_env.action),
                                      False,
                                      name='last_mean')
        self._last_logstd = tf.Variable(tf.zeros_like(self._batch_env.action),
                                        False,
                                        name='last_logstd')
    self._penalty = tf.Variable(self._config.kl_init_penalty, False, dtype=tf.float32)
    self._optimizer = self._config.optimizer(self._config.learning_rate)

  def begin_episode(self, agent_indices):
    """Reset the recurrent states and stored episode.

    Args:
      agent_indices: Tensor containing current batch indices.

    Returns:
      Summary tensor.
    """
    with tf.name_scope('begin_episode/'):
      if self._last_state is None:
        reset_state = tf.no_op()
      else:
        reset_state = utility.reinit_nested_vars(self._last_state, agent_indices)
      reset_buffer = self._episodes.clear(agent_indices)
      with tf.control_dependencies([reset_state, reset_buffer]):
        return tf.constant('')

  def perform(self, agent_indices, observ):
    """Compute batch of actions and a summary for a batch of observation.

    Args:
      agent_indices: Tensor containing current batch indices.
      observ: Tensor of a batch of observations for all agents.

    Returns:
      Tuple of action batch tensor and summary tensor.
    """
    with tf.name_scope('perform/'):
      observ = self._observ_filter.transform(observ)
      if self._last_state is None:
        state = None
      else:
        state = tf.contrib.framework.nest.map_structure(lambda x: tf.gather(x, agent_indices),
                                                        self._last_state)
      output = self._network(observ[:, None], tf.ones(observ.shape[0]), state)
      action = tf.cond(self._is_training, output.policy.sample, lambda: output.mean)
      logprob = output.policy.log_prob(action)[:, 0]
      # pylint: disable=g-long-lambda
      summary = tf.cond(
          self._should_log, lambda: tf.summary.merge([
              tf.summary.histogram('mean', output.mean[:, 0]),
              tf.summary.histogram('std', tf.exp(output.logstd[:, 0])),
              tf.summary.histogram('action', action[:, 0]),
              tf.summary.histogram('logprob', logprob)
          ]), str)
      # Remember current policy to append to memory in the experience callback.
      if self._last_state is None:
        assign_state = tf.no_op()
      else:
        assign_state = utility.assign_nested_vars(self._last_state, output.state, agent_indices)
      with tf.control_dependencies([
          assign_state,
          tf.scatter_update(self._last_action, agent_indices, action[:, 0]),
          tf.scatter_update(self._last_mean, agent_indices, output.mean[:, 0]),
          tf.scatter_update(self._last_logstd, agent_indices, output.logstd[:, 0])
      ]):
        return tf.check_numerics(action[:, 0], 'action'), tf.identity(summary)

  def experience(self, agent_indices, observ, action, reward, unused_done, unused_nextob):
    """Process the transition tuple of the current step.

    When training, add the current transition tuple to the memory and update
    the streaming statistics for observations and rewards. A summary string is
    returned if requested at this step.

    Args:
      agent_indices: Tensor containing current batch indices.
      observ: Batch tensor of observations.
      action: Batch tensor of actions.
      reward: Batch tensor of rewards.
      unused_done: Batch tensor of done flags.
      unused_nextob: Batch tensor of successor observations.

    Returns:
      Summary tensor.
    """
    with tf.name_scope('experience/'):
      return tf.cond(
          self._is_training,
          # pylint: disable=g-long-lambda
          lambda: self._define_experience(agent_indices, observ, action, reward),
          str)

  def _define_experience(self, agent_indices, observ, action, reward):
    """Implement the branch of experience() entered during training."""
    update_filters = tf.summary.merge(
        [self._observ_filter.update(observ),
         self._reward_filter.update(reward)])
    with tf.control_dependencies([update_filters]):
      if self._config.train_on_agent_action:
        # NOTE: Doesn't seem to change much.
        action = self._last_action
      batch = (observ, action, tf.gather(self._last_mean,
                                         agent_indices), tf.gather(self._last_logstd,
                                                                   agent_indices), reward)
      append = self._episodes.append(batch, agent_indices)
    with tf.control_dependencies([append]):
      norm_observ = self._observ_filter.transform(observ)
      norm_reward = tf.reduce_mean(self._reward_filter.transform(reward))
      # pylint: disable=g-long-lambda
      summary = tf.cond(
          self._should_log, lambda: tf.summary.merge([
              update_filters,
              self._observ_filter.summary(),
              self._reward_filter.summary(),
              tf.summary.scalar('memory_size', self._memory_index),
              tf.summary.histogram('normalized_observ', norm_observ),
              tf.summary.histogram('action', self._last_action),
              tf.summary.scalar('normalized_reward', norm_reward)
          ]), str)
      return summary

  def end_episode(self, agent_indices):
    """Add episodes to the memory and perform update steps if memory is full.

    During training, add the collected episodes of the batch indices that
    finished their episode to the memory. If the memory is full, train on it,
    and then clear the memory. A summary string is returned if requested at
    this step.

    Args:
      agent_indices: Tensor containing current batch indices.

    Returns:
       Summary tensor.
    """
    with tf.name_scope('end_episode/'):
      return tf.cond(self._is_training, lambda: self._define_end_episode(agent_indices), str)

  def _define_end_episode(self, agent_indices):
    """Implement the branch of end_episode() entered during training."""
    episodes, length = self._episodes.data(agent_indices)
    space_left = self._config.update_every - self._memory_index
    use_episodes = tf.range(tf.minimum(tf.shape(agent_indices)[0], space_left))
    episodes = [tf.gather(elem, use_episodes) for elem in episodes]
    append = self._memory.replace(episodes, tf.gather(length, use_episodes),
                                  use_episodes + self._memory_index)
    with tf.control_dependencies([append]):
      inc_index = self._memory_index.assign_add(tf.shape(use_episodes)[0])
    with tf.control_dependencies([inc_index]):
      memory_full = self._memory_index >= self._config.update_every
      return tf.cond(memory_full, self._training, str)

  def _training(self):
    """Perform multiple training iterations of both policy and value baseline.

    Training on the episodes collected in the memory. Reset the memory
    afterwards. Always returns a summary string.

    Returns:
      Summary tensor.
    """
    with tf.name_scope('training'):
      assert_full = tf.assert_equal(self._memory_index, self._config.update_every)
      with tf.control_dependencies([assert_full]):
        data = self._memory.data()
      (observ, action, old_mean, old_logstd, reward), length = data
      with tf.control_dependencies([tf.assert_greater(length, 0)]):
        length = tf.identity(length)
      observ = self._observ_filter.transform(observ)
      reward = self._reward_filter.transform(reward)
      update_summary = self._perform_update_steps(observ, action, old_mean, old_logstd, reward,
                                                  length)
      with tf.control_dependencies([update_summary]):
        penalty_summary = self._adjust_penalty(observ, old_mean, old_logstd, length)
      with tf.control_dependencies([penalty_summary]):
        clear_memory = tf.group(self._memory.clear(), self._memory_index.assign(0))
      with tf.control_dependencies([clear_memory]):
        weight_summary = utility.variable_summaries(tf.trainable_variables(),
                                                    self._config.weight_summaries)
        return tf.summary.merge([update_summary, penalty_summary, weight_summary])

  def _perform_update_steps(self, observ, action, old_mean, old_logstd, reward, length):
    """Perform multiple update steps of value function and policy.

    The advantage is computed once at the beginning and shared across
    iterations. We need to decide for the summary of one iteration, and thus
    choose the one after half of the iterations.

    Args:
      observ: Sequences of observations.
      action: Sequences of actions.
      old_mean: Sequences of action means of the behavioral policy.
      old_logstd: Sequences of action log stddevs of the behavioral policy.
      reward: Sequences of rewards.
      length: Batch of sequence lengths.

    Returns:
      Summary tensor.
    """
    return_ = utility.discounted_return(reward, length, self._config.discount)
    value = self._network(observ, length).value
    if self._config.gae_lambda:
      advantage = utility.lambda_return(reward, value, length, self._config.discount,
                                        self._config.gae_lambda)
    else:
      advantage = return_ - value
    mean, variance = tf.nn.moments(advantage, axes=[0, 1], keep_dims=True)
    advantage = (advantage - mean) / (tf.sqrt(variance) + 1e-8)
    advantage = tf.Print(advantage,
                         [tf.reduce_mean(return_), tf.reduce_mean(value)], 'return and value: ')
    advantage = tf.Print(advantage, [tf.reduce_mean(advantage)], 'normalized advantage: ')
    # pylint: disable=g-long-lambda
    value_loss, policy_loss, summary = tf.scan(lambda _1, _2: self._update_step(
        observ, action, old_mean, old_logstd, reward, advantage, length),
                                               tf.range(self._config.update_epochs), [0., 0., ''],
                                               parallel_iterations=1)
    print_losses = tf.group(tf.Print(0, [tf.reduce_mean(value_loss)], 'value loss: '),
                            tf.Print(0, [tf.reduce_mean(policy_loss)], 'policy loss: '))
    with tf.control_dependencies([value_loss, policy_loss, print_losses]):
      return summary[self._config.update_epochs // 2]

  def _update_step(self, observ, action, old_mean, old_logstd, reward, advantage, length):
    """Compute the current combined loss and perform a gradient update step.

    Args:
      observ: Sequences of observations.
      action: Sequences of actions.
      old_mean: Sequences of action means of the behavioral policy.
      old_logstd: Sequences of action log stddevs of the behavioral policy.
      reward: Sequences of reward.
      advantage: Sequences of advantages.
      length: Batch of sequence lengths.

    Returns:
      Tuple of value loss, policy loss, and summary tensor.
    """
    value_loss, value_summary = self._value_loss(observ, reward, length)
    network = self._network(observ, length)
    policy_loss, policy_summary = self._policy_loss(network.mean, network.logstd, old_mean,
                                                    old_logstd, action, advantage, length)
    value_gradients, value_variables = (zip(*self._optimizer.compute_gradients(value_loss)))
    policy_gradients, policy_variables = (zip(*self._optimizer.compute_gradients(policy_loss)))
    all_gradients = value_gradients + policy_gradients
    all_variables = value_variables + policy_variables
    optimize = self._optimizer.apply_gradients(zip(all_gradients, all_variables))
    summary = tf.summary.merge([
        value_summary, policy_summary,
        tf.summary.scalar('value_gradient_norm', tf.global_norm(value_gradients)),
        tf.summary.scalar('policy_gradient_norm', tf.global_norm(policy_gradients)),
        utility.gradient_summaries(zip(value_gradients, value_variables), dict(value=r'.*')),
        utility.gradient_summaries(zip(policy_gradients, policy_variables), dict(policy=r'.*'))
    ])
    with tf.control_dependencies([optimize]):
      return [tf.identity(x) for x in (value_loss, policy_loss, summary)]

  def _value_loss(self, observ, reward, length):
    """Compute the loss function for the value baseline.

    The value loss is the difference between empirical and approximated returns
    over the collected episodes. Returns the loss tensor and a summary strin.

    Args:
      observ: Sequences of observations.
      reward: Sequences of reward.
      length: Batch of sequence lengths.

    Returns:
      Tuple of loss tensor and summary tensor.
    """
    with tf.name_scope('value_loss'):
      value = self._network(observ, length).value
      return_ = utility.discounted_return(reward, length, self._config.discount)
      advantage = return_ - value
      value_loss = 0.5 * self._mask(advantage**2, length)
      summary = tf.summary.merge([
          tf.summary.histogram('value_loss', value_loss),
          tf.summary.scalar('avg_value_loss', tf.reduce_mean(value_loss))
      ])
      value_loss = tf.reduce_mean(value_loss)
      return tf.check_numerics(value_loss, 'value_loss'), summary

  def _policy_loss(self, mean, logstd, old_mean, old_logstd, action, advantage, length):
    """Compute the policy loss composed of multiple components.

    1. The policy gradient loss is importance sampled from the data-collecting
       policy at the beginning of training.
    2. The second term is a KL penalty between the policy at the beginning of
       training and the current policy.
    3. Additionally, if this KL already changed more than twice the target
       amount, we activate a strong penalty discouraging further divergence.

    Args:
      mean: Sequences of action means of the current policy.
      logstd: Sequences of action log stddevs of the current policy.
      old_mean: Sequences of action means of the behavioral policy.
      old_logstd: Sequences of action log stddevs of the behavioral policy.
      action: Sequences of actions.
      advantage: Sequences of advantages.
      length: Batch of sequence lengths.

    Returns:
      Tuple of loss tensor and summary tensor.
    """
    with tf.name_scope('policy_loss'):
      entropy = utility.diag_normal_entropy(mean, logstd)
      kl = tf.reduce_mean(
          self._mask(utility.diag_normal_kl(old_mean, old_logstd, mean, logstd), length), 1)
      policy_gradient = tf.exp(
          utility.diag_normal_logpdf(mean, logstd, action) -
          utility.diag_normal_logpdf(old_mean, old_logstd, action))
      surrogate_loss = -tf.reduce_mean(
          self._mask(policy_gradient * tf.stop_gradient(advantage), length), 1)
      kl_penalty = self._penalty * kl
      cutoff_threshold = self._config.kl_target * self._config.kl_cutoff_factor
      cutoff_count = tf.reduce_sum(tf.cast(kl > cutoff_threshold, tf.int32))
      with tf.control_dependencies(
          [tf.cond(cutoff_count > 0, lambda: tf.Print(0, [cutoff_count], 'kl cutoff! '), int)]):
        kl_cutoff = (self._config.kl_cutoff_coef * tf.cast(kl > cutoff_threshold, tf.float32) *
                     (kl - cutoff_threshold)**2)
      policy_loss = surrogate_loss + kl_penalty + kl_cutoff
      summary = tf.summary.merge([
          tf.summary.histogram('entropy', entropy),
          tf.summary.histogram('kl', kl),
          tf.summary.histogram('surrogate_loss', surrogate_loss),
          tf.summary.histogram('kl_penalty', kl_penalty),
          tf.summary.histogram('kl_cutoff', kl_cutoff),
          tf.summary.histogram('kl_penalty_combined', kl_penalty + kl_cutoff),
          tf.summary.histogram('policy_loss', policy_loss),
          tf.summary.scalar('avg_surr_loss', tf.reduce_mean(surrogate_loss)),
          tf.summary.scalar('avg_kl_penalty', tf.reduce_mean(kl_penalty)),
          tf.summary.scalar('avg_policy_loss', tf.reduce_mean(policy_loss))
      ])
      policy_loss = tf.reduce_mean(policy_loss, 0)
      return tf.check_numerics(policy_loss, 'policy_loss'), summary

  def _adjust_penalty(self, observ, old_mean, old_logstd, length):
    """Adjust the KL policy between the behavioral and current policy.

    Compute how much the policy actually changed during the multiple
    update steps. Adjust the penalty strength for the next training phase if we
    overshot or undershot the target divergence too much.

    Args:
      observ: Sequences of observations.
      old_mean: Sequences of action means of the behavioral policy.
      old_logstd: Sequences of action log stddevs of the behavioral policy.
      length: Batch of sequence lengths.

    Returns:
      Summary tensor.
    """
    with tf.name_scope('adjust_penalty'):
      network = self._network(observ, length)
      assert_change = tf.assert_equal(tf.reduce_all(tf.equal(network.mean, old_mean)),
                                      False,
                                      message='policy should change')
      print_penalty = tf.Print(0, [self._penalty], 'current penalty: ')
      with tf.control_dependencies([assert_change, print_penalty]):
        kl_change = tf.reduce_mean(
            self._mask(utility.diag_normal_kl(old_mean, old_logstd, network.mean, network.logstd),
                       length))
        kl_change = tf.Print(kl_change, [kl_change], 'kl change: ')
        maybe_increase = tf.cond(
            kl_change > 1.3 * self._config.kl_target,
            # pylint: disable=g-long-lambda
            lambda: tf.Print(self._penalty.assign(self._penalty * 1.5), [0], 'increase penalty '),
            float)
        maybe_decrease = tf.cond(
            kl_change < 0.7 * self._config.kl_target,
            # pylint: disable=g-long-lambda
            lambda: tf.Print(self._penalty.assign(self._penalty / 1.5), [0], 'decrease penalty '),
            float)
      with tf.control_dependencies([maybe_increase, maybe_decrease]):
        return tf.summary.merge([
            tf.summary.scalar('kl_change', kl_change),
            tf.summary.scalar('penalty', self._penalty)
        ])

  def _mask(self, tensor, length):
    """Set padding elements of a batch of sequences to zero.

    Useful to then safely sum along the time dimension.

    Args:
      tensor: Tensor of sequences.
      length: Batch of sequence lengths.

    Returns:
      Masked sequences.
    """
    with tf.name_scope('mask'):
      range_ = tf.range(tensor.shape[1].value)
      mask = tf.cast(range_[None, :] < length[:, None], tf.float32)
      masked = tensor * mask
      return tf.check_numerics(masked, 'masked')



"""The leg controller class interface."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
from typing import Any


class LegController(object):  # pytype: disable=ignored-metaclass
  """Generates the leg control signal."""

  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def reset(self, current_time: float):
    """Resets the controller's internal state."""
    pass

  @abc.abstractmethod
  def update(self, current_time: float):
    """Updates the controller's internal state."""
    pass

  @abc.abstractmethod
  def get_action(self) -> Any:
    """Gets the control signal e.g. torques/positions for the leg."""
    pass



"""A model based controller framework."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
from typing import Any, Callable

from pybullet_envs.minitaur.agents.baseline_controller import gait_generator as gait_generator_lib
from pybullet_envs.minitaur.agents.baseline_controller import leg_controller as leg_controller_lib
from pybullet_envs.minitaur.agents.baseline_controller import state_estimator as state_estimator_lib


class LocomotionController(object):
  """Generates the quadruped locomotion.

  The actual effect of this controller depends on the composition of each
  individual subcomponent.

  """

  def __init__(
      self,
      robot: Any,
      gait_generator: gait_generator_lib.GaitGenerator,
      state_estimator: state_estimator_lib.StateEstimatorBase,
      swing_leg_controller: leg_controller_lib.LegController,
      stance_leg_controller: leg_controller_lib.LegController,
      clock: Callable[[], float] = None,
  ):
    """Initializes the class.

    Args:
      robot: A robot instance.
      gait_generator: Generates the leg swing/stance pattern.
      state_estimator: Estimates the state of the robot (e.g. center of mass
        position or velocity that may not be observable from sensors).
      swing_leg_controller: Generates motor actions for swing legs.
      stance_leg_controller: Generates motor actions for stance legs.
      clock: A real or fake clock source.
    """
    self._robot = robot
    self._clock = clock
    if clock is None:
      self._clock = time.time
    self._reset_time = self._clock()
    self._time_since_reset = 0
    self._gait_generator = gait_generator
    self._state_estimator = state_estimator
    self._swing_leg_controller = swing_leg_controller
    self._stance_leg_controller = stance_leg_controller

  @property
  def swing_leg_controller(self):
    return self._swing_leg_controller

  @property
  def stance_leg_controller(self):
    return self._stance_leg_controller

  @property
  def gait_generator(self):
    return self._gait_generator

  @property
  def state_estimator(self):
    return self._state_estimator

  def reset(self):
    self._reset_time = self._clock()
    self._time_since_reset = 0
    self._gait_generator.reset(self._time_since_reset)
    self._state_estimator.reset(self._time_since_reset)
    self._swing_leg_controller.reset(self._time_since_reset)
    self._stance_leg_controller.reset(self._time_since_reset)

  def update(self):
    self._time_since_reset = self._clock() - self._reset_time
    self._gait_generator.update(self._time_since_reset)
    self._state_estimator.update(self._time_since_reset)
    self._swing_leg_controller.update(self._time_since_reset)
    self._stance_leg_controller.update(self._time_since_reset)

  def get_action(self):
    """Returns the control ouputs (e.g. positions/torques) for all motors."""
    swing_action = self._swing_leg_controller.get_action()
    stance_action = self._stance_leg_controller.get_action()
    action = []
    for joint_id in range(self._robot.num_motors):
      if joint_id in swing_action:
        action.extend(swing_action[joint_id])
      else:
        assert joint_id in stance_action
        action.extend(stance_action[joint_id])
    return action



"""A wrapper for pybullet to manage different clients."""
from __future__ import absolute_import
from __future__ import division
import os
import functools
import inspect
import pybullet


class BulletClient(object):
  """A wrapper for pybullet to manage different clients."""

  def __init__(self, connection_mode=None, hostName=None, options=''):
    """Creates a Bullet client and connects to a simulation.

    Args:
      connection_mode:
        `None` connects to an existing simulation or, if fails, creates a
          new headless simulation,
        `pybullet.GUI` creates a new simulation with a GUI,
        `pybullet.DIRECT` creates a headless simulation,
        `pybullet.SHARED_MEMORY` connects to an existing simulation.
    """
    self._shapes = {}
    self._pid = os.getpid()
    if connection_mode is None:
      self._client = pybullet.connect(pybullet.SHARED_MEMORY, options=options)
      if self._client >= 0:
        return
      else:
        connection_mode = pybullet.DIRECT
    if hostName is None:
        self._client = pybullet.connect(connection_mode, options=options)
    else:
        self._client = pybullet.connect(connection_mode, hostName=hostName, options=options)

  def __del__(self):
    """Clean up connection if not already done."""
    if self._client>=0 and self._pid == os.getpid():
      try:
        pybullet.disconnect(physicsClientId=self._client)
        self._client = -1
      except pybullet.error:
        pass

  def __getattr__(self, name):
    """Inject the client id into Bullet functions."""
    attribute = getattr(pybullet, name)
    if inspect.isbuiltin(attribute):
      attribute = functools.partial(attribute, physicsClientId=self._client)
    if name=="disconnect":
      self._client = -1 
    return attribute



"""Adds random forces to the base of Minitaur during the simulation steps."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
parentdir = os.path.dirname(os.path.dirname(parentdir))
os.sys.path.insert(0, parentdir)

import math
import numpy as np
from pybullet_envs.minitaur.envs import env_randomizer_base

_PERTURBATION_START_STEP = 100
_PERTURBATION_INTERVAL_STEPS = 200
_PERTURBATION_DURATION_STEPS = 10
_HORIZONTAL_FORCE_UPPER_BOUND = 120
_HORIZONTAL_FORCE_LOWER_BOUND = 240
_VERTICAL_FORCE_UPPER_BOUND = 300
_VERTICAL_FORCE_LOWER_BOUND = 500


class MinitaurPushRandomizer(env_randomizer_base.EnvRandomizerBase):
  """Applies a random impulse to the base of Minitaur."""

  def __init__(
      self,
      perturbation_start_step=_PERTURBATION_START_STEP,
      perturbation_interval_steps=_PERTURBATION_INTERVAL_STEPS,
      perturbation_duration_steps=_PERTURBATION_DURATION_STEPS,
      horizontal_force_bound=None,
      vertical_force_bound=None,
  ):
    """Initializes the randomizer.

    Args:
      perturbation_start_step: No perturbation force before the env has advanced
        this amount of steps.
      perturbation_interval_steps: The step interval between applying
        perturbation forces.
      perturbation_duration_steps: The duration of the perturbation force.
      horizontal_force_bound: The lower and upper bound of the applied force
        magnitude when projected in the horizontal plane.
      vertical_force_bound: The z component (abs value) bound of the applied
        perturbation force.
    """
    self._perturbation_start_step = perturbation_start_step
    self._perturbation_interval_steps = perturbation_interval_steps
    self._perturbation_duration_steps = perturbation_duration_steps
    self._horizontal_force_bound = (horizontal_force_bound if horizontal_force_bound else
                                    [_HORIZONTAL_FORCE_LOWER_BOUND, _HORIZONTAL_FORCE_UPPER_BOUND])
    self._vertical_force_bound = (vertical_force_bound if vertical_force_bound else
                                  [_VERTICAL_FORCE_LOWER_BOUND, _VERTICAL_FORCE_UPPER_BOUND])

  def randomize_env(self, env):
    """Randomizes the simulation environment.

    Args:
      env: The Minitaur gym environment to be randomized.
    """
    pass

  def randomize_step(self, env):
    """Randomizes simulation steps.

    Will be called at every timestep. May add random forces/torques to Minitaur.

    Args:
      env: The Minitaur gym environment to be randomized.
    """
    base_link_ids = env.minitaur.chassis_link_ids
    if env.env_step_counter % self._perturbation_interval_steps == 0:
      self._applied_link_id = base_link_ids[np.random.randint(0, len(base_link_ids))]
      horizontal_force_magnitude = np.random.uniform(self._horizontal_force_bound[0],
                                                     self._horizontal_force_bound[1])
      theta = np.random.uniform(0, 2 * math.pi)
      vertical_force_magnitude = np.random.uniform(self._vertical_force_bound[0],
                                                   self._vertical_force_bound[1])
      self._applied_force = horizontal_force_magnitude * np.array(
          [math.cos(theta), math.sin(theta), 0]) + np.array([0, 0, -vertical_force_magnitude])

    if (env.env_step_counter % self._perturbation_interval_steps <
        self._perturbation_duration_steps) and (env.env_step_counter >=
                                                self._perturbation_start_step):
      env.pybullet_client.applyExternalForce(objectUniqueId=env.minitaur.quadruped,
                                             linkIndex=self._applied_link_id,
                                             forceObj=self._applied_force,
                                             posObj=[0.0, 0.0, 0.0],
                                             flags=env.pybullet_client.LINK_FRAME)



"""This file implements an accurate motor model."""

import numpy as np
VOLTAGE_CLIPPING = 50
# TODO(b/73728631): Clamp the pwm signal instead of the OBSERVED_TORQUE_LIMIT.
OBSERVED_TORQUE_LIMIT = 5.7
MOTOR_VOLTAGE = 16.0
MOTOR_RESISTANCE = 0.186
MOTOR_TORQUE_CONSTANT = 0.0954
MOTOR_VISCOUS_DAMPING = 0
MOTOR_SPEED_LIMIT = MOTOR_VOLTAGE / (MOTOR_VISCOUS_DAMPING + MOTOR_TORQUE_CONSTANT)
NUM_MOTORS = 8


class MotorModel(object):
  """The accurate motor model, which is based on the physics of DC motors.

  The motor model support two types of control: position control and torque
  control. In position control mode, a desired motor angle is specified, and a
  torque is computed based on the internal motor model. When the torque control
  is specified, a pwm signal in the range of [-1.0, 1.0] is converted to the
  torque.

  The internal motor model takes the following factors into consideration:
  pd gains, viscous friction, back-EMF voltage and current-torque profile.
  """

  def __init__(self, torque_control_enabled=False, kp=1.2, kd=0):
    self._torque_control_enabled = torque_control_enabled
    self._kp = kp
    self._kd = kd
    self._resistance = MOTOR_RESISTANCE
    self._voltage = MOTOR_VOLTAGE
    self._torque_constant = MOTOR_TORQUE_CONSTANT
    self._viscous_damping = MOTOR_VISCOUS_DAMPING
    self._current_table = [0, 10, 20, 30, 40, 50, 60]
    self._torque_table = [0, 1, 1.9, 2.45, 3.0, 3.25, 3.5]
    self._strength_ratios = [1.0] * NUM_MOTORS

  def set_strength_ratios(self, ratios):
    """Set the strength of each motors relative to the default value.

    Args:
      ratios: The relative strength of motor output. A numpy array ranging from
        0.0 to 1.0.
    """
    self._strength_ratios = np.array(ratios)

  def set_motor_gains(self, kp, kd):
    """Set the gains of all motors.

    These gains are PD gains for motor positional control. kp is the
    proportional gain and kd is the derivative gain.

    Args:
      kp: proportional gain of the motors.
      kd: derivative gain of the motors.
    """
    self._kp = kp
    self._kd = kd

  def set_voltage(self, voltage):
    self._voltage = voltage

  def get_voltage(self):
    return self._voltage

  def set_viscous_damping(self, viscous_damping):
    self._viscous_damping = viscous_damping

  def get_viscous_dampling(self):
    return self._viscous_damping

  def convert_to_torque(self,
                        motor_commands,
                        motor_angle,
                        motor_velocity,
                        true_motor_velocity,
                        kp=None,
                        kd=None):
    """Convert the commands (position control or torque control) to torque.

    Args:
      motor_commands: The desired motor angle if the motor is in position
        control mode. The pwm signal if the motor is in torque control mode.
      motor_angle: The motor angle observed at the current time step. It is
        actually the true motor angle observed a few milliseconds ago (pd
        latency).
      motor_velocity: The motor velocity observed at the current time step, it
        is actually the true motor velocity a few milliseconds ago (pd latency).
      true_motor_velocity: The true motor velocity. The true velocity is used
        to compute back EMF voltage and viscous damping.
      kp: Proportional gains for the motors' PD controllers. If not provided, it
        uses the default kp of the minitaur for all the motors.
      kd: Derivative gains for the motors' PD controllers. If not provided, it
        uses the default kp of the minitaur for all the motors.

    Returns:
      actual_torque: The torque that needs to be applied to the motor.
      observed_torque: The torque observed by the sensor.
    """
    if self._torque_control_enabled:
      pwm = motor_commands
    else:
      if kp is None:
        kp = np.full(NUM_MOTORS, self._kp)
      if kd is None:
        kd = np.full(NUM_MOTORS, self._kd)
      pwm = -1 * kp * (motor_angle - motor_commands) - kd * motor_velocity

    pwm = np.clip(pwm, -1.0, 1.0)
    return self._convert_to_torque_from_pwm(pwm, true_motor_velocity)

  def _convert_to_torque_from_pwm(self, pwm, true_motor_velocity):
    """Convert the pwm signal to torque.

    Args:
      pwm: The pulse width modulation.
      true_motor_velocity: The true motor velocity at the current moment. It is
        used to compute the back EMF voltage and the viscous damping.
    Returns:
      actual_torque: The torque that needs to be applied to the motor.
      observed_torque: The torque observed by the sensor.
    """
    observed_torque = np.clip(
        self._torque_constant * (np.asarray(pwm) * self._voltage / self._resistance),
        -OBSERVED_TORQUE_LIMIT, OBSERVED_TORQUE_LIMIT)

    # Net voltage is clipped at 50V by diodes on the motor controller.
    voltage_net = np.clip(
        np.asarray(pwm) * self._voltage -
        (self._torque_constant + self._viscous_damping) * np.asarray(true_motor_velocity),
        -VOLTAGE_CLIPPING, VOLTAGE_CLIPPING)
    current = voltage_net / self._resistance
    current_sign = np.sign(current)
    current_magnitude = np.absolute(current)
    # Saturate torque based on empirical current relation.
    actual_torque = np.interp(current_magnitude, self._current_table, self._torque_table)
    actual_torque = np.multiply(current_sign, actual_torque)
    actual_torque = np.multiply(self._strength_ratios, actual_torque)
    return actual_torque, observed_torque



"""A wrapped LocomotionGymEnv with functions that change the world and task."""

import numbers
from typing import Optional, Sequence, Text, Tuple, Union

import numpy as np

import gin


@gin.configurable
class CurriculumParameter(object):
  """Definition of a env parameter tuned by a curriculum."""

  def __init__(self,
               name: Text,
               init_val: numbers.Real,
               bounds: Tuple[numbers.Real, numbers.Real]):
    """A parameter to tune throughout the curriculum.

    Args:
      name: The name of the curriculum parameter. This must be the name of
        attribute of the scene class.
      init_val: The value to use at the start of the curriculum.
      bounds: A tuple of [lower_bound, upper_bound] defining the minimum and
        maximum values of the parameter.
    """
    self.name = name

    if not isinstance(bounds[0], type(bounds[1])):
      raise ValueError("All elements in [bounds] must be of the same type.")

    if (init_val < min(bounds)) or (init_val > max(bounds)):
      raise ValueError("Initial parameter value must lie in range defined"
                       " by [bounds].")

    if not isinstance(init_val, type(bounds[0])):
      raise ValueError("[init_val] type must match the type of the elements"
                       " in [bounds].")

    self.init_val = init_val
    self.bounds = bounds
    self.dtype = type(init_val)


@gin.configurable
class RandomSamplingCurriculumParameter(CurriculumParameter):
  """Env parameter whose value is sampled randomly without curriculum."""

  def __init__(self, name: Text, bounds: Tuple[numbers.Real, numbers.Real]):
    """A parameter whose value to sample randomly.

    Args:
      name: The name of the parameter.
      bounds: A tuple of [lower_bound, upper_bound] defining the minimum and
        maximum values of the parameter.
    """
    super(RandomSamplingCurriculumParameter, self).__init__(
        name=name, init_val=bounds[0], bounds=bounds)

  def sample(
      self, step: numbers.Real,
      curriculum_steps: Optional[numbers.Real] = None) -> Union[int, float]:
    del step, curriculum_steps
    sampled_val = np.random.uniform(*self.bounds)

    if not isinstance(self.bounds[0], float):
      sampled_val = int(round(sampled_val))

    return sampled_val

  def __call__(
      self, step: numbers.Real,
      curriculum_steps: Optional[numbers.Real] = None) -> Union[int, float]:
    return self.sample(step, curriculum_steps)


@gin.configurable
class LinearStepBasedCurriculumParameter(CurriculumParameter):
  """Definition of a env parameter tuned by a linear time-based curriculum."""

  def __init__(self,
               name: Text,
               init_val: numbers.Real,
               bounds: Tuple[numbers.Real, numbers.Real],
               curriculum_steps: Optional[numbers.Real] = None):
    """A parameter to tune throughout the curriculum.

    Args:
      name: The name of the curriculum parameter. This must be the name of
        attribute of the scene class.
      init_val: The value to use at the start of the curriculum.
      bounds: A tuple of [lower_bound, upper_bound] defining the minimum and
        maximum values of the parameter.
      curriculum_steps: Integer defining the number of steps to take when
        varying the curriculum parameter value from the init_val to either
        bound. If None is specified, then the curriculum must provide the
        curriculum_steps when sampling.
    """
    super(LinearStepBasedCurriculumParameter, self).__init__(
        name=name, init_val=init_val, bounds=bounds)
    self.curriculum_steps = curriculum_steps

  def get_bounds_at_step(
      self, step: numbers.Real,
      curriculum_steps: Optional[numbers.Real] = None
    ) -> Tuple[Union[int, float], Union[int, float]]:
    """Compute the bounds of the parameter at the current step.

    Args:
      step: An integer defining the current timestep.
      curriculum_steps: Optional curriculum steps. Must be passed if not passed
        during initialization.
    Returns:
      A tuple containing the lower bound and upper bound at the current step.
    """

    if not self.curriculum_steps and not curriculum_steps:
      raise ValueError("curriculum_steps not defined. Must be passed upon"
                       " initialization or must specified by curriculum"
                       " wrapper env on method call.")

    if self.curriculum_steps:
      curriculum_steps = self.curriculum_steps

    prog = min(float(step) / curriculum_steps, 1.0)
    curr_lower_bound = self.init_val - prog * (self.init_val - self.bounds[0])
    curr_upper_bound = self.init_val + prog * (self.bounds[1] - self.init_val)

    if not isinstance(self.bounds[0], float):
      curr_lower_bound = int(round(curr_lower_bound))
      curr_upper_bound = int(round(curr_upper_bound))

    return (curr_lower_bound, curr_upper_bound)

  def sample(
      self, step: numbers.Real,
      curriculum_steps: Optional[numbers.Real] = None) -> Union[int, float]:
    sampled_val = np.random.uniform(*self.get_bounds_at_step(
        step, curriculum_steps))

    if not isinstance(self.bounds[0], float):
      sampled_val = int(round(sampled_val))

    return sampled_val

  def __call__(
      self, step: numbers.Real,
      curriculum_steps: Optional[numbers.Real] = None) -> Union[int, float]:
    return self.sample(step, curriculum_steps)


@gin.configurable
class Task(object):
  """Defines a single task in the environment and its corresponding params."""

  def __init__(self, name: Text,
               curriculum_parameters: Sequence[CurriculumParameter]):
    """Initialize the task.

    Args:
      name: The name of the task.
      curriculum_parameters: A list of CurriculumParameter instances which
        define the parameters that this task makes use of.
    """
    self.name = name
    self.curriculum_parameters = curriculum_parameters


@gin.configurable
class StepBasedCurriculumWrapperEnv(object):
  """A wrapper to tune the scene parameters linearly with the steps taken."""

  def __init__(self, env, tasks: Sequence[Task],
               default_curriculum_steps: Optional[numbers.Real] = None,
               reset_total_step_count_val: numbers.Real = -1,
               steps_before_curriculum_start: numbers.Real = 0):
    """Initializes the linear curriculum wrapper env.

    Args:
      env: An instance of a (potentially previously wrapped) LocomotionGymEnv.
      tasks: Various tasks to shuffle through throughout the curriculum.
      default_curriculum_steps: Optional default value for curriculum steps.
      reset_total_step_count_val: Step at which to reset the total step count.
        The internal total_step_count is reset to 0 once this value is reached.
      steps_before_curriculum_start: Steps to take in environment before the
        curriculum begins.
    """
    self._gym_env = env
    self._tasks = tasks
    self._default_curriculum_steps = default_curriculum_steps
    self._reset_total_step_count_val = reset_total_step_count_val
    self._steps_before_curriculum_start = steps_before_curriculum_start

    # Total number of environment steps.
    self._total_step_count = 0

  def __getattr__(self, attr):
    return getattr(self._gym_env, attr)

  def set_scene_params(self):
    # Choose a task at random.
    curr_task = np.random.choice(self._tasks)

    # Cycle through the task's parameters and set them in the scene.
    self._gym_env.scene.reset_scene_params()
    self._gym_env.scene.scene_type = curr_task.name
    for curriculum_parameter in curr_task.curriculum_parameters:
      setattr(
          self._gym_env.scene, curriculum_parameter.name,
          curriculum_parameter(
              self._total_step_count - self._steps_before_curriculum_start,
              self._default_curriculum_steps))

  def reset(self, *args, **kwargs):
    """Reset and adjust the environment."""

    # Update the total step count.
    self._total_step_count += self._gym_env.env_step_counter
    if self._reset_total_step_count_val >= 0:
      if self._total_step_count >= self._reset_total_step_count_val:
        self._total_step_count = 0

    if self._total_step_count < self._steps_before_curriculum_start:
      return self._get_observation()

    self.set_scene_params()
    self._gym_env.reset(*args, **kwargs)

    return self._get_observation()




import time
import numpy as np
import math

useNullSpace = 0
useDynamics = 1
useIKFast = 0 #ikfast doesn't get solutions and is actually slower than pybullet IK (300us versus 150us), probably a configuration issue
if useIKFast:
  import ikfastpy

ikSolver = 0
xarmEndEffectorIndex = 6
xarmNumDofs = 6

ll = [-17]*xarmNumDofs
#upper limits for null space (todo: set them to proper range)
ul = [17]*xarmNumDofs
#joint ranges for null space (todo: set them to proper range)
jr = [17]*xarmNumDofs
#restposes for null space
jointPositions=[0,0,0,0,0,0]
rp = jointPositions
jointPoses = jointPositions
class XArm6Sim(object):
  def __init__(self, bullet_client, offset):
    if useIKFast:
      # Initialize kinematics for UR5 robot arm
      self.xarm_kin = ikfastpy.PyKinematics()
      self.n_joints = self.xarm_kin.getDOF()
      print("numJoints IKFast=", self.n_joints)

    self.bullet_client = bullet_client
    self.offset = np.array(offset)
    self.jointPoses = [0]*xarmNumDofs
    #print("offset=",offset)
    flags = self.bullet_client.URDF_ENABLE_CACHED_GRAPHICS_SHAPES
    legos=[]
    self.bullet_client.loadURDF("tray/traybox.urdf", [0.4+offset[0], offset[1], offset[2]], [0,0,0,1], flags=flags)
    legos.append(self.bullet_client.loadURDF("lego/lego.urdf",np.array([0.3, 0.1, 0.3])+self.offset, flags=flags))
    legos.append(self.bullet_client.loadURDF("lego/lego.urdf",np.array([0.3, -0.1, 0.3])+self.offset, flags=flags))
    legos.append(self.bullet_client.loadURDF("lego/lego.urdf",np.array([0.5, 0.1, 0.3])+self.offset, flags=flags))
    sphereId = self.bullet_client.loadURDF("sphere_small.urdf",np.array( [0.4, 0, 0.3])+self.offset, flags=flags)
    self.bullet_client.loadURDF("sphere_small.urdf",np.array( [0.3, 0, 0.3])+self.offset, flags=flags)
    self.bullet_client.loadURDF("sphere_small.urdf",np.array( [0.5, 0, 0.3])+self.offset, flags=flags)
    orn=[0,0,0,1]
    self.xarm = self.bullet_client.loadURDF("xarm/xarm6_robot.urdf", np.array([0,0,0])+self.offset, orn, useFixedBase=True, flags=flags)
    index = 0
    for j in range(self.bullet_client.getNumJoints(self.xarm)):
      self.bullet_client.changeDynamics(self.xarm, j, linearDamping=0, angularDamping=0)
      info = self.bullet_client.getJointInfo(self.xarm, j)
  
      jointName = info[1]
      jointType = info[2]
      if (jointType == self.bullet_client.JOINT_PRISMATIC):
        
        self.bullet_client.resetJointState(self.xarm, j, jointPositions[index]) 
        index=index+1
      if (jointType == self.bullet_client.JOINT_REVOLUTE):
        self.bullet_client.resetJointState(self.xarm, j, jointPositions[index]) 
        index=index+1
    self.t = 0.
  def reset(self):
    pass

  def step(self):
    t = self.t
    self.t += 1./60.
    
    pos = [0.407+self.offset[0]+0.2 * math.sin(1.5 * t), 0.0+self.offset[1]+0.2 * math.cos(1.5 * t), 0.12+self.offset[2]]
    #orn = self.bullet_client.getQuaternionFromEuler([math.pi/2.,0.,0.])
    orn = [1,0,0,0]
    
    if useIKFast:
      
      #mat = self.bullet_client.getMatrixFromQuaternion(orn)
      #ee_pose = np.array([mat[0],mat[1],mat[2],pos[0],
      #          mat[3],mat[4],mat[5],pos[1],
      #          mat[6],mat[7],mat[8],pos[2]])
      #ee_pose = np.asarray(ee_pose).reshape(3,4)
      joint_angles = [0.39,0,0,0,0,0]#-3.1,-1.6,1.6,-1.6,-1.6,0.] # in radians
      self.bullet_client.submitProfileTiming("ikfast.forwardDynamics")
      ee_pose = self.xarm_kin.forward(joint_angles)
      self.bullet_client.submitProfileTiming()
      ee_pose = np.asarray(ee_pose).reshape(3,4) 
      #print("ee_pose=",ee_pose)
      self.bullet_client.submitProfileTiming("ikfast.inverseDynamics")
      joint_configs = self.xarm_kin.inverse(ee_pose.reshape(-1).tolist())
      self.bullet_client.submitProfileTiming()
      n_solutions = int(len(joint_configs)/self.n_joints)
      print("%d solutions found:"%(n_solutions))
      joint_configs = np.asarray(joint_configs).reshape(n_solutions,self.n_joints)
      for joint_config in joint_configs:
        print(joint_config)
      jointPoses = [0]*self.n_joints
      
    else: 
      if useNullSpace:
        jointPoses = self.bullet_client.calculateInverseKinematics(self.xarm,xarmEndEffectorIndex, pos, orn,lowerLimits=ll, 
          upperLimits=ul,jointRanges=jr, restPoses=np.array(self.jointPoses).tolist(),residualThreshold=1e-5, maxNumIterations=50)
        self.jointPoses = [0,0,jointPoses[2],jointPoses[3],jointPoses[4],jointPoses[5]]
      else:
        jointPoses = self.bullet_client.calculateInverseKinematics(self.xarm,xarmEndEffectorIndex, pos, orn, maxNumIterations=50)
    #print("jointPoses=",jointPoses)
    if useDynamics:
      for i in range(xarmNumDofs):
          self.bullet_client.setJointMotorControl2(self.xarm, i+1, self.bullet_client.POSITION_CONTROL, jointPoses[i],force=5 * 240.)
    else:
      for i in range(xarmNumDofs):
        self.bullet_client.resetJointState(self.xarm, i+1, jointPoses[i])
    ls = self.bullet_client.getLinkState(self.xarm, xarmEndEffectorIndex, computeForwardKinematics=True)
    linkComPos=np.array(ls[0])
    linkComOrn=ls[1]
    linkUrdfPos=np.array(ls[4])
    linkUrdfOrn=ls[5]
    #print("linkComPos=",linkComPos)
    #print("linkUrdfOrn=",linkUrdfOrn)
    mat = self.bullet_client.getMatrixFromQuaternion(linkUrdfOrn)
    #print("mat=",mat)
    self.bullet_client.addUserDebugLine(pos, linkUrdfPos, [1,0,0,1],lifeTime=100)
    diff = linkUrdfPos-np.array(pos)
    #print("diff=",diff)
    
  



"""A proto buffer based logging system for minitaur experiments.

The logging system records the time since reset, base position, orientation,
angular velocity and motor information (joint angle, speed, and torque) into a
proto buffer. See minitaur_logging.proto for more details. The episode_proto is
updated per time step by the environment and saved onto disk for each episode.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(os.path.dirname(currentdir))
os.sys.path.insert(0, parentdir)

import datetime
import os
import time
import logging

from pybullet_envs.minitaur.envs import minitaur_logging_pb2

NUM_MOTORS = 8


def _update_base_state(base_state, values):
  base_state.x = values[0]
  base_state.y = values[1]
  base_state.z = values[2]


def preallocate_episode_proto(episode_proto, max_num_steps):
  """Preallocate the memory for proto buffer.

  Dynamically allocating memory as the protobuf expands causes unexpected delay
  that is not tolerable with locomotion control.

  Args:
    episode_proto: The proto that holds the state/action data for the current
      episode.
    max_num_steps: The max number of steps that will be recorded in the proto.
      The state/data over max_num_steps will not be stored in the proto.
  """
  for _ in range(max_num_steps):
    step_log = episode_proto.state_action.add()
    step_log.info_valid = False
    step_log.time.seconds = 0
    step_log.time.nanos = 0
    for _ in range(NUM_MOTORS):
      motor_state = step_log.motor_states.add()
      motor_state.angle = 0
      motor_state.velocity = 0
      motor_state.torque = 0
      motor_state.action = 0
    _update_base_state(step_log.base_position, [0, 0, 0])
    _update_base_state(step_log.base_orientation, [0, 0, 0])
    _update_base_state(step_log.base_angular_vel, [0, 0, 0])


def update_episode_proto(episode_proto, minitaur, action, step):
  """Update the episode proto by appending the states/action of the minitaur.

  Note that the state/data over max_num_steps preallocated
  (len(episode_proto.state_action)) will not be stored in the proto.
  Args:
    episode_proto: The proto that holds the state/action data for the current
      episode.
    minitaur: The minitaur instance. See envs.minitaur for details.
    action: The action applied at this time step. The action is an 8-element
      numpy floating-point array.
    step: The current step index.
  """
  max_num_steps = len(episode_proto.state_action)
  if step >= max_num_steps:
    logging.warning("{}th step is not recorded in the logging since only {} steps were "
                       "pre-allocated.".format(step, max_num_steps))
    return
  step_log = episode_proto.state_action[step]
  step_log.info_valid = minitaur.IsObservationValid()
  time_in_seconds = minitaur.GetTimeSinceReset()
  step_log.time.seconds = int(time_in_seconds)
  step_log.time.nanos = int((time_in_seconds - int(time_in_seconds)) * 1e9)

  motor_angles = minitaur.GetMotorAngles()
  motor_velocities = minitaur.GetMotorVelocities()
  motor_torques = minitaur.GetMotorTorques()
  for i in range(minitaur.num_motors):
    step_log.motor_states[i].angle = motor_angles[i]
    step_log.motor_states[i].velocity = motor_velocities[i]
    step_log.motor_states[i].torque = motor_torques[i]
    step_log.motor_states[i].action = action[i]

  _update_base_state(step_log.base_position, minitaur.GetBasePosition())
  _update_base_state(step_log.base_orientation, minitaur.GetBaseRollPitchYaw())
  _update_base_state(step_log.base_angular_vel, minitaur.GetBaseRollPitchYawRate())


class MinitaurLogging(object):
  """A logging system that records the states/action of the minitaur."""

  def __init__(self, log_path=None):
    self._log_path = log_path

  # TODO(jietan): Consider using recordio to write the logs.
  def save_episode(self, episode_proto):
    """Save episode_proto to self._log_path.

    self._log_path is the directory name. A time stamp is the file name of the
    log file. For example, when self._log_path is "/tmp/logs/", the actual
    log file would be "/tmp/logs/yyyy-mm-dd-hh:mm:ss".

    Args:
      episode_proto: The proto that holds the states/action for the current
        episode that needs to be save to disk.
    Returns:
      The full log path, including the directory name and the file name.
    """
    if not self._log_path or not episode_proto.state_action:
      return self._log_path
    if not os.path.exists(self._log_path):
      os.mkdir(self._log_path)
    ts = time.time()
    time_stamp = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d-%H%M%S")
    log_path = os.path.join(self._log_path, "minitaur_log_{}".format(time_stamp))
    
    with open(log_path, 'wb') as f:
      f.write(episode_proto.SerializeToString())
    return log_path

  def restore_episode(self, log_path):
    """Restore the episodic proto from the log path.

    Args:
      log_path: The full path of the log file.
    Returns:
      The minitaur episode proto.
    """
    with open(log_path, 'rb') as f:
      content = f.read()
      episode_proto = minitaur_logging_pb2.MinitaurEpisode()
      episode_proto.ParseFromString(content)
      return episode_proto



"""Wraps a quadruped walking controller for navigation control."""

from typing import Any
import gin
from gym import spaces
import numpy as np

from pybullet_envs.minitaur.agents.baseline_controller import com_velocity_estimator
from pybullet_envs.minitaur.agents.baseline_controller import locomotion_controller
from pybullet_envs.minitaur.agents.baseline_controller import openloop_gait_generator
from pybullet_envs.minitaur.agents.baseline_controller import raibert_swing_leg_controller
from pybullet_envs.minitaur.agents.baseline_controller import torque_stance_leg_controller

_N_LEGS = 4
_STANCE_DURATION_SECONDS = [
    0.25
] * _N_LEGS  # The stance phase duration for each leg.
_DUTY_FACTOR = [
    0.6
] * _N_LEGS  # Percentage of the leg in the stance phase within the cycle.
_BODY_HEIGHT = 0.45
_INIT_PHASE_FULL_CYCLE = [0, 0.5, 0.5, 0]


def _setup_controller(robot: Any) -> locomotion_controller.LocomotionController:
  """Creates the controller."""
  desired_speed = (0, 0)
  desired_twisting_speed = 0

  gait_generator = openloop_gait_generator.OpenloopGaitGenerator(
      robot,
      stance_duration=_STANCE_DURATION_SECONDS,
      duty_factor=_DUTY_FACTOR,
      initial_leg_phase=_INIT_PHASE_FULL_CYCLE)
  state_estimator = com_velocity_estimator.COMVelocityEstimator(robot)
  sw_controller = raibert_swing_leg_controller.RaibertSwingLegController(
      robot,
      gait_generator,
      state_estimator,
      desired_speed=desired_speed,
      desired_twisting_speed=desired_twisting_speed,
      desired_height=_BODY_HEIGHT,
  )
  st_controller = torque_stance_leg_controller.TorqueStanceLegController(
      robot,
      gait_generator,
      state_estimator,
      desired_speed=desired_speed,
      desired_twisting_speed=desired_twisting_speed,
      desired_body_height=_BODY_HEIGHT,
      body_mass=215 / 9.8,
      body_inertia=(0.07335, 0, 0, 0, 0.25068, 0, 0, 0, 0.25447),
  )

  controller = locomotion_controller.LocomotionController(
      robot=robot,
      gait_generator=gait_generator,
      state_estimator=state_estimator,
      swing_leg_controller=sw_controller,
      stance_leg_controller=st_controller,
      clock=robot.GetTimeSinceReset)
  return controller


def _update_controller_params(
    controller: locomotion_controller.LocomotionController,
    lin_speed: np.ndarray, ang_speed: float):
  """Apply the desired speed and twisting speed."""
  controller.swing_leg_controller.desired_speed = lin_speed
  controller.swing_leg_controller.desired_twisting_speed = ang_speed
  controller.stance_leg_controller.desired_speed = lin_speed
  controller.stance_leg_controller.desired_twisting_speed = ang_speed


@gin.configurable
class WalkingWrapper(object):
  """Wraps a baseline walking controller for Laikago/Vision60."""

  def __init__(self,
               gym_env: Any,
               action_repeat=20,
               speed_bound=(-0.3, 0.3),
               angular_speed_bound=(-0.3, 0.3)):
    """Initialzes the wrapped env.

    Args:
      gym_env: An instance of LocomotionGymEnv.
      action_repeat: Number of control steps to run low level controller with
        the high level inputs per step().
      speed_bound: The min/max of the input speed.
      angular_speed_bound: The min/max of the twisting speed.
    """
    self._gym_env = gym_env
    self._controller = _setup_controller(self._gym_env.robot)
    self._action_repeat = action_repeat

    action_low = np.array(speed_bound)
    action_high = np.array(angular_speed_bound)

    # Overwrite the action space.
    self.action_space = spaces.Box(action_low, action_high)

  def reset(self, *args, **kwargs) -> Any:
    obs = self._gym_env.reset(*args, **kwargs)
    # The robot instance might have been replaced if hard_reset is called. We
    # just recreate the controller.
    self._controller = _setup_controller(self._gym_env.robot)
    self._controller.reset()
    return obs

  def __getattr__(self, attr):
    return getattr(self._gym_env, attr)

  def step(self, action) -> Any:
    """Steps the wrapped env with high level commands.

    Args:
      action: A high level command containing the desired linear and angular
        speed of the robot base. The speed can be adjusted at any time.

    Returns:
      The gym env observation, reward, termination, and additional info.

    """
    lin_speed = np.array((action[0], 0, 0))
    ang_speed = action[1]
    _update_controller_params(self._controller, lin_speed, ang_speed)
    for _ in range(self._action_repeat):
      self._controller.update()
      hybrid_action = self._controller.get_action()
      obs, reward, done, info = self._gym_env.step(hybrid_action)
      if done:
        break
    return obs, reward, done, info



# Lint as: python3
"""Defines the laikago robot related constants and URDF specs."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import gin

URDF_PATH = "laikago/laikago_toes_zup.urdf"

NUM_MOTORS = 12
NUM_LEGS = 4
MOTORS_PER_LEG = 3

INIT_RACK_POSITION = [0, 0, 1]
INIT_POSITION = [0, 0, 0.48]

# Will be default to (0, 0, 0, 1) once the new laikago_toes_zup.urdf checked in.
INIT_ORIENTATION = [0, 0, 0, 1]

# Can be different from the motors, although for laikago they are the same list.
JOINT_NAMES = (
    # front right leg
    "FR_hip_motor_2_chassis_joint",
    "FR_upper_leg_2_hip_motor_joint",
    "FR_lower_leg_2_upper_leg_joint",
    # front left leg
    "FL_hip_motor_2_chassis_joint",
    "FL_upper_leg_2_hip_motor_joint",
    "FL_lower_leg_2_upper_leg_joint",
    # rear right leg
    "RR_hip_motor_2_chassis_joint",
    "RR_upper_leg_2_hip_motor_joint",
    "RR_lower_leg_2_upper_leg_joint",
    # rear left leg
    "RL_hip_motor_2_chassis_joint",
    "RL_upper_leg_2_hip_motor_joint",
    "RL_lower_leg_2_upper_leg_joint",
)

INIT_ABDUCTION_ANGLE = 0
INIT_HIP_ANGLE = 0.67
INIT_KNEE_ANGLE = -1.25

# Note this matches the Laikago SDK/control convention, but is different from
# URDF's internal joint angles which needs to be computed using the joint
# offsets and directions. The conversion formula is (sdk_joint_angle + offset) *
# joint direction.
INIT_JOINT_ANGLES = collections.OrderedDict(
    zip(JOINT_NAMES,
        (INIT_ABDUCTION_ANGLE, INIT_HIP_ANGLE, INIT_KNEE_ANGLE) * NUM_LEGS))

# Used to convert the robot SDK joint angles to URDF joint angles.
JOINT_DIRECTIONS = collections.OrderedDict(
    zip(JOINT_NAMES, (-1, 1, 1, 1, 1, 1, -1, 1, 1, 1, 1, 1)))

HIP_JOINT_OFFSET = 0.0
UPPER_LEG_JOINT_OFFSET = -0.6
KNEE_JOINT_OFFSET = 0.66

# Used to convert the robot SDK joint angles to URDF joint angles.
JOINT_OFFSETS = collections.OrderedDict(
    zip(JOINT_NAMES,
        [HIP_JOINT_OFFSET, UPPER_LEG_JOINT_OFFSET, KNEE_JOINT_OFFSET] *
        NUM_LEGS))

LEG_NAMES = (
    "front_right",
    "front_left",
    "rear_right",
    "rear_left",
)

LEG_ORDER = (
    "front_right",
    "front_left",
    "back_right",
    "back_left",
)

END_EFFECTOR_NAMES = (
    "jtoeFR",
    "jtoeFL",
    "jtoeRR",
    "jtoeRL",
)

MOTOR_NAMES = JOINT_NAMES
MOTOR_GROUP = collections.OrderedDict((
    (LEG_NAMES[0], JOINT_NAMES[0:3]),
    (LEG_NAMES[1], JOINT_NAMES[3:6]),
    (LEG_NAMES[2], JOINT_NAMES[6:9]),
    (LEG_NAMES[3], JOINT_NAMES[9:12]),
))

# Regulates the joint angle change when in position control mode.
MAX_MOTOR_ANGLE_CHANGE_PER_STEP = 0.12

# The hip joint location in the CoM frame.
HIP_POSITIONS = collections.OrderedDict((
    (LEG_NAMES[0], (0.21, -0.1157, 0)),
    (LEG_NAMES[1], (0.21, 0.1157, 0)),
    (LEG_NAMES[2], (-0.21, -0.1157, 0)),
    (LEG_NAMES[3], (-0.21, 0.1157, 0)),
))

# Add the gin constants to be used for gin binding in config. Append "LAIKAGO_"
# for unique binding names.
gin.constant("laikago_constants.LAIKAGO_NUM_MOTORS", NUM_MOTORS)
gin.constant("laikago_constants.LAIKAGO_URDF_PATH", URDF_PATH)
gin.constant("laikago_constants.LAIKAGO_INIT_POSITION", INIT_POSITION)
gin.constant("laikago_constants.LAIKAGO_INIT_ORIENTATION", INIT_ORIENTATION)
gin.constant("laikago_constants.LAIKAGO_INIT_JOINT_ANGLES", INIT_JOINT_ANGLES)
gin.constant("laikago_constants.LAIKAGO_JOINT_DIRECTIONS", JOINT_DIRECTIONS)
gin.constant("laikago_constants.LAIKAGO_JOINT_OFFSETS", JOINT_OFFSETS)
gin.constant("laikago_constants.LAIKAGO_MOTOR_NAMES", MOTOR_NAMES)
gin.constant("laikago_constants.LAIKAGO_END_EFFECTOR_NAMES", END_EFFECTOR_NAMES)
gin.constant("laikago_constants.LAIKAGO_MOTOR_GROUP", MOTOR_GROUP)



import numpy as np
import copy
from pybullet_utils.logger import Logger
import inspect as inspect
from pybullet_envs.deep_mimic.env.env import Env
import pybullet_utils.math_util as MathUtil


class ReplayBuffer(object):
  TERMINATE_KEY = 'terminate'
  PATH_START_KEY = 'path_start'
  PATH_END_KEY = 'path_end'

  def __init__(self, buffer_size):
    assert buffer_size > 0

    self.buffer_size = buffer_size
    self.total_count = 0
    self.buffer_head = 0
    self.buffer_tail = MathUtil.INVALID_IDX
    self.num_paths = 0
    self._sample_buffers = dict()
    self.buffers = None

    self.clear()
    return

  def sample(self, n):
    curr_size = self.get_current_size()
    assert curr_size > 0

    idx = np.empty(n, dtype=int)
    # makes sure that the end states are not sampled
    for i in range(n):
      while True:
        curr_idx = np.random.randint(0, curr_size, size=1)[0]
        curr_idx += self.buffer_tail
        curr_idx = np.mod(curr_idx, self.buffer_size)

        if not self.is_path_end(curr_idx):
          break
      idx[i] = curr_idx

    return idx

  def sample_filtered(self, n, key):
    assert key in self._sample_buffers
    curr_buffer = self._sample_buffers[key]
    idx = curr_buffer.sample(n)
    return idx

  def count_filtered(self, key):
    curr_buffer = self._sample_buffers[key]
    return curr_buffer.count

  def get(self, key, idx):
    return self.buffers[key][idx]

  def get_all(self, key):
    return self.buffers[key]

  def get_idx_filtered(self, key):
    assert key in self._sample_buffers
    curr_buffer = self._sample_buffers[key]
    idx = curr_buffer.slot_to_idx[:curr_buffer.count]
    return idx

  def get_path_start(self, idx):
    return self.buffers[self.PATH_START_KEY][idx]

  def get_path_end(self, idx):
    return self.buffers[self.PATH_END_KEY][idx]

  def get_pathlen(self, idx):
    is_array = isinstance(idx, np.ndarray) or isinstance(idx, list)
    if not is_array:
      idx = [idx]

    n = len(idx)
    start_idx = self.get_path_start(idx)
    end_idx = self.get_path_end(idx)
    pathlen = np.empty(n, dtype=int)

    for i in range(n):
      curr_start = start_idx[i]
      curr_end = end_idx[i]
      if curr_start < curr_end:
        curr_len = curr_end - curr_start
      else:
        curr_len = self.buffer_size - curr_start + curr_end
      pathlen[i] = curr_len

    if not is_array:
      pathlen = pathlen[0]

    return pathlen

  def is_valid_path(self, idx):
    start_idx = self.get_path_start(idx)
    valid = start_idx != MathUtil.INVALID_IDX
    return valid

  def store(self, path):
    start_idx = MathUtil.INVALID_IDX
    n = path.pathlength()

    if (n > 0):
      assert path.is_valid()

      if path.check_vals():
        if self.buffers is None:
          self._init_buffers(path)

        idx = self._request_idx(n + 1)
        self._store_path(path, idx)
        self._add_sample_buffers(idx)

        self.num_paths += 1
        self.total_count += n + 1
        start_idx = idx[0]
      else:
        Logger.print2('Invalid path data value detected')

    return start_idx

  def clear(self):
    self.buffer_head = 0
    self.buffer_tail = MathUtil.INVALID_IDX
    self.num_paths = 0

    for key in self._sample_buffers:
      self._sample_buffers[key].clear()
    return

  def get_next_idx(self, idx):
    next_idx = np.mod(idx + 1, self.buffer_size)
    return next_idx

  def is_terminal_state(self, idx):
    terminate_flags = self.buffers[self.TERMINATE_KEY][idx]
    terminate = terminate_flags != Env.Terminate.Null.value
    is_end = self.is_path_end(idx)
    terminal_state = np.logical_and(terminate, is_end)
    return terminal_state

  def check_terminal_flag(self, idx, flag):
    terminate_flags = self.buffers[self.TERMINATE_KEY][idx]
    terminate = terminate_flags == flag.value
    return terminate

  def is_path_end(self, idx):
    is_end = self.buffers[self.PATH_END_KEY][idx] == idx
    return is_end

  def add_filter_key(self, key):
    assert self.get_current_size() == 0
    if key not in self._sample_buffers:
      self._sample_buffers[key] = SampleBuffer(self.buffer_size)
    return

  def get_current_size(self):
    if self.buffer_tail == MathUtil.INVALID_IDX:
      return 0
    elif self.buffer_tail < self.buffer_head:
      return self.buffer_head - self.buffer_tail
    else:
      return self.buffer_size - self.buffer_tail + self.buffer_head

  def _check_flags(self, key, flags):
    return (flags & key) == key

  def _add_sample_buffers(self, idx):
    flags = self.buffers['flags']
    for key in self._sample_buffers:
      curr_buffer = self._sample_buffers[key]
      filter_idx = [
          i for i in idx if (self._check_flags(key, flags[i]) and not self.is_path_end(i))
      ]
      curr_buffer.add(filter_idx)
    return

  def _free_sample_buffers(self, idx):
    for key in self._sample_buffers:
      curr_buffer = self._sample_buffers[key]
      curr_buffer.free(idx)
    return

  def _init_buffers(self, path):
    self.buffers = dict()
    self.buffers[self.PATH_START_KEY] = MathUtil.INVALID_IDX * np.ones(self.buffer_size, dtype=int)
    self.buffers[self.PATH_END_KEY] = MathUtil.INVALID_IDX * np.ones(self.buffer_size, dtype=int)

    for key in dir(path):
      val = getattr(path, key)
      if not key.startswith('__') and not inspect.ismethod(val):
        if key == self.TERMINATE_KEY:
          self.buffers[self.TERMINATE_KEY] = np.zeros(shape=[self.buffer_size], dtype=int)
        else:
          val_type = type(val[0])
          is_array = val_type == np.ndarray
          if is_array:
            shape = [self.buffer_size, val[0].shape[0]]
            dtype = val[0].dtype
          else:
            shape = [self.buffer_size]
            dtype = val_type

          self.buffers[key] = np.zeros(shape, dtype=dtype)
    return

  def _request_idx(self, n):
    assert n + 1 < self.buffer_size  # bad things can happen if path is too long

    remainder = n
    idx = []

    start_idx = self.buffer_head
    while remainder > 0:
      end_idx = np.minimum(start_idx + remainder, self.buffer_size)
      remainder -= (end_idx - start_idx)

      free_idx = list(range(start_idx, end_idx))
      self._free_idx(free_idx)
      idx += free_idx
      start_idx = 0

    self.buffer_head = (self.buffer_head + n) % self.buffer_size
    return idx

  def _free_idx(self, idx):
    assert (idx[0] <= idx[-1])
    n = len(idx)
    if self.buffer_tail != MathUtil.INVALID_IDX:
      update_tail = idx[0] <= idx[-1] and idx[0] <= self.buffer_tail and idx[-1] >= self.buffer_tail
      update_tail |= idx[0] > idx[-1] and (idx[0] <= self.buffer_tail or
                                           idx[-1] >= self.buffer_tail)

      if update_tail:
        i = 0
        while i < n:
          curr_idx = idx[i]
          if self.is_valid_path(curr_idx):
            start_idx = self.get_path_start(curr_idx)
            end_idx = self.get_path_end(curr_idx)
            pathlen = self.get_pathlen(curr_idx)

            if start_idx < end_idx:
              self.buffers[self.PATH_START_KEY][start_idx:end_idx + 1] = MathUtil.INVALID_IDX
              self._free_sample_buffers(list(range(start_idx, end_idx + 1)))
            else:
              self.buffers[self.PATH_START_KEY][start_idx:self.buffer_size] = MathUtil.INVALID_IDX
              self.buffers[self.PATH_START_KEY][0:end_idx + 1] = MathUtil.INVALID_IDX
              self._free_sample_buffers(list(range(start_idx, self.buffer_size)))
              self._free_sample_buffers(list(range(0, end_idx + 1)))

            self.num_paths -= 1
            i += pathlen + 1
            self.buffer_tail = (end_idx + 1) % self.buffer_size
          else:
            i += 1
    else:
      self.buffer_tail = idx[0]
    return

  def _store_path(self, path, idx):
    n = path.pathlength()
    for key, data in self.buffers.items():
      if key != self.PATH_START_KEY and key != self.PATH_END_KEY and key != self.TERMINATE_KEY:
        val = getattr(path, key)
        val_len = len(val)
        assert val_len == n or val_len == n + 1
        data[idx[:val_len]] = val

    self.buffers[self.TERMINATE_KEY][idx] = path.terminate.value
    self.buffers[self.PATH_START_KEY][idx] = idx[0]
    self.buffers[self.PATH_END_KEY][idx] = idx[-1]
    return


class SampleBuffer(object):

  def __init__(self, size):
    self.idx_to_slot = np.empty(shape=[size], dtype=int)
    self.slot_to_idx = np.empty(shape=[size], dtype=int)
    self.count = 0
    self.clear()
    return

  def clear(self):
    self.idx_to_slot.fill(MathUtil.INVALID_IDX)
    self.slot_to_idx.fill(MathUtil.INVALID_IDX)
    self.count = 0
    return

  def is_valid(self, idx):
    return self.idx_to_slot[idx] != MathUtil.INVALID_IDX

  def get_size(self):
    return self.idx_to_slot.shape[0]

  def add(self, idx):
    for i in idx:
      if not self.is_valid(i):
        new_slot = self.count
        assert new_slot >= 0

        self.idx_to_slot[i] = new_slot
        self.slot_to_idx[new_slot] = i
        self.count += 1
    return

  def free(self, idx):
    for i in idx:
      if self.is_valid(i):
        slot = self.idx_to_slot[i]
        last_slot = self.count - 1
        last_idx = self.slot_to_idx[last_slot]

        self.idx_to_slot[last_idx] = slot
        self.slot_to_idx[slot] = last_idx
        self.idx_to_slot[i] = MathUtil.INVALID_IDX
        self.slot_to_idx[last_slot] = MathUtil.INVALID_IDX
        self.count -= 1
    return

  def sample(self, n):
    if self.count > 0:
      slots = np.random.randint(0, self.count, size=n)
      idx = self.slot_to_idx[slots]
    else:
      idx = np.empty(shape=[0], dtype=int)
    return idx

  def check_consistency(self):
    valid = True
    if self.count < 0:
      valid = False

    if valid:
      for i in range(self.get_size()):
        if self.is_valid(i):
          s = self.idx_to_slot[i]
          if self.slot_to_idx[s] != i:
            valid = False
            break

        s2i = self.slot_to_idx[i]
        if s2i != MathUtil.INVALID_IDX:
          i2s = self.idx_to_slot[s2i]
          if i2s != i:
            valid = False
            break

    count0 = np.sum(self.idx_to_slot == MathUtil.INVALID_IDX)
    count1 = np.sum(self.slot_to_idx == MathUtil.INVALID_IDX)
    valid &= count0 == count1
    return valid



"""An example to run the gym environment that a minitaur follows a ball.

"""

import math

import tf.compat.v1 as tf
import minitaur_ball_gym_env


def FollowBallManualPolicy():
  """An example of a minitaur following a ball."""
  env = minitaur_ball_gym_env.MinitaurBallGymEnv(render=True,
                                                 pd_control_enabled=True,
                                                 on_rack=False)
  observation = env.reset()
  sum_reward = 0
  steps = 100000
  for _ in range(steps):
    action = [math.tanh(observation[0] * 4)]
    observation, reward, done, _ = env.step(action)
    sum_reward += reward
    if done:
      tf.logging.info("Return is {}".format(sum_reward))
      observation = env.reset()
      sum_reward = 0


def main():
  FollowBallManualPolicy()


if __name__ == '__main__':
  main()



# Lint as: python3
"""The abstract robot class."""

import abc
from typing import Optional, Sequence

# Action names for robots operating kinematically.
LINEAR_VELOCITY = "linear_velocity"
ANGULAR_VELOCITY = "angular_velocity"


class RobotBase(metaclass=abc.ABCMeta):
  """The base class for all robots used in the mobility team."""

  @abc.abstractmethod
  def reset(
      self,
      base_position: Optional[Sequence[float]] = None,
      base_orientation_quaternion: Optional[Sequence[float]] = None) -> None:
    """Resets the states (e.g. pose and sensor readings) of the robot.

    This is called at the start of each episode by the environment.

    Args:
      base_position: Robot base position after reset. If None, robot stay where
        it was after reset. For robot that does not support reset with position
        change, a ValueError should be raised.
      base_orientation_quaternion: Robot base orientation after reset. If None,
        robot stays in pre-reset orientation. For robot that does not support
        reset with orientation change, a ValueError should be raised.
    """
    pass

  @abc.abstractmethod
  def terminate(self):
    """Shuts down the robot."""
    pass

  @abc.abstractmethod
  def pre_control_step(self, action):
    """Processes the input action before the action repeat loop.

    We assume that an action sent to the real robot is sticky, i.e. it will be
    executed until a new action is received after some time. To simulate this,
    we introduced the action_repeat parameter, to reflect how many time steps it
    takes for the policy to generate a new action. That is, for each control
    step, the simulation contains an inner loop:

      robot.pre_control_step(action)  # Smooth or interpolate the action
      for i in range(action_repeat):
        robot.apply_action(action)
        bullet.stepSimulation(time_step)  # Step the sim for one time step
        robot.receive_observation()  # Update the sensor observations
      robot.post_control_step() # Update some internal variables.

    Args:
      action: Data type depends on the robot. Can be desired motor
        position/torques for legged robots, or desired velocity/angular velocity
        for wheeled robots.
    """
    pass

  @abc.abstractmethod
  def apply_action(self, action):
    """Applies the action to the robot."""
    pass

  @abc.abstractmethod
  def receive_observation(self):
    """Updates the robot sensor readings."""
    pass

  @abc.abstractmethod
  def post_control_step(self):
    """Updates some internal variables such as step counters."""
    pass

  @property
  def action_space(self):
    """The action spec of the robot."""
    raise NotImplementedError("action_space is not implemented")

  @property
  @abc.abstractmethod
  def action_names(self):
    """Name of each action in the action_space.

    This is a structure of strings with the same shape as the action space,
    where each string describes the corresponding element of the action space
    (for example, a kinematic robot might return ("linear_velocity",
    "angular_velocity")). Used for logging in the safety layer.
    """

  @property
  def sensors(self):
    """Returns the sensors on this robot.

    Sensors are the main interface between the robot class and the gym
    environment. Sensors can return what the robot can measure (e.g.
    joint angles, IMU readings), and can represent more general quantities, i.e.
    the last action taken, that can be part of the observation space.
    Sensor classes are used by the robot class to the specify its observation
    space.

    """
    raise NotImplementedError("sensors property not implemented")

  @property
  def base_orientation_quaternion(self):
    """Returns the base pose as a quaternion in format (x, y, z, w).

    These properties differ from the sensor interfaces, as they represent
    the built-in measurable quantities. We assume most robots have an IMU at
    its base to measure the base pose. Actually, some sensor classes like the
    base pose sensor and joint angle sensor will call these built-in methods. In
    general, how these quantities can be extracted depends on the specific real
    robots.

    """
    raise NotImplementedError("base_orientation_quaternion is not implemented")

  @property
  def base_roll_pitch_yaw(self):
    """Returns the base roll, pitch, and yaw angles."""
    raise NotImplementedError("base_roll_pitch_yaw is not implemented")

  @property
  def base_roll_pitch_yaw_rate(self):
    raise NotImplementedError("base_roll_pitch_yaw_rate is not implemented")

  @property
  def base_position(self):
    raise NotImplementedError("base_position is not implemented")



from robot_bases import MJCFBasedRobot
import numpy as np


class InvertedPendulum(MJCFBasedRobot):
  swingup = False

  def __init__(self):
    MJCFBasedRobot.__init__(self, 'inverted_pendulum.xml', 'cart', action_dim=1, obs_dim=5)

  def robot_specific_reset(self, bullet_client):
    self._p = bullet_client
    self.pole = self.parts["pole"]
    self.slider = self.jdict["slider"]
    self.j1 = self.jdict["hinge"]
    u = self.np_random.uniform(low=-.1, high=.1)
    self.j1.reset_current_position(u if not self.swingup else 3.1415 + u, 0)
    self.j1.set_motor_torque(0)

  def apply_action(self, a):
    assert (np.isfinite(a).all())
    if not np.isfinite(a).all():
      print("a is inf")
      a[0] = 0
    self.slider.set_motor_torque(100 * float(np.clip(a[0], -1, +1)))

  def calc_state(self):
    self.theta, theta_dot = self.j1.current_position()
    x, vx = self.slider.current_position()
    assert (np.isfinite(x))

    if not np.isfinite(x):
      print("x is inf")
      x = 0

    if not np.isfinite(vx):
      print("vx is inf")
      vx = 0

    if not np.isfinite(self.theta):
      print("theta is inf")
      self.theta = 0

    if not np.isfinite(theta_dot):
      print("theta_dot is inf")
      theta_dot = 0

    return np.array([x, vx, np.cos(self.theta), np.sin(self.theta), theta_dot])


class InvertedPendulumSwingup(InvertedPendulum):
  swingup = True


class InvertedDoublePendulum(MJCFBasedRobot):

  def __init__(self):
    MJCFBasedRobot.__init__(self, 'inverted_double_pendulum.xml', 'cart', action_dim=1, obs_dim=9)

  def robot_specific_reset(self, bullet_client):
    self._p = bullet_client
    self.pole2 = self.parts["pole2"]
    self.slider = self.jdict["slider"]
    self.j1 = self.jdict["hinge"]
    self.j2 = self.jdict["hinge2"]
    u = self.np_random.uniform(low=-.1, high=.1, size=[2])
    self.j1.reset_current_position(float(u[0]), 0)
    self.j2.reset_current_position(float(u[1]), 0)
    self.j1.set_motor_torque(0)
    self.j2.set_motor_torque(0)

  def apply_action(self, a):
    assert (np.isfinite(a).all())
    self.slider.set_motor_torque(200 * float(np.clip(a[0], -1, +1)))

  def calc_state(self):
    theta, theta_dot = self.j1.current_position()
    gamma, gamma_dot = self.j2.current_position()
    x, vx = self.slider.current_position()
    self.pos_x, _, self.pos_y = self.pole2.pose().xyz()
    assert (np.isfinite(x))
    return np.array([
        x,
        vx,
        self.pos_x,
        np.cos(theta),
        np.sin(theta),
        theta_dot,
        np.cos(gamma),
        np.sin(gamma),
        gamma_dot,
    ])
